'''
Random structure generation using PyXtal (https://github.com/qzhu2017/PyXtal)
'''
import collections
from contextlib import redirect_stdout, redirect_stderr
from io import StringIO
from logging import getLogger
from multiprocessing import Process, Queue
import random
import sys

import numpy as np
from pymatgen.core import Structure
from pymatgen.core.periodic_table import DummySpecie
from pyxtal import pyxtal
from pyxtal.tolerance import Tol_matrix

from ...util.struc_util import check_distance, sort_by_atype, get_nat
from ...util.struc_util import get_atype_dummy, scale_cell_mol, rot_mat


logger = getLogger('cryspy')


def gen_struc(
        nstruc,
        atype,
        nat,
        mindist,
        spgnum,
        symprec=0.01,
        id_offset=0,
        vol_factor=1.1,
        vol_mu=None,
        vol_sigma=None,
        vc=False,
        ll_nat=None,
        ul_nat=None
    ):
    '''
    Generate random structures for given space groups

        tuple may be replaced by list

    # ---------- args
    nstruc (int): number of structures to be generated
    atype (tuple): atom type (e.g. ('Na', 'Cl'))
    nat (tuple): number of atoms (e.g. (4, 4)), None if vc=True
    mindist (): minimum interatomic distance (e.g. ((2.0, 1.5), (1.5, 2.0)))
    spgnum (str, int, or tuple): space group number 'all', 0, or tuple of space group numbers
    symprec (float): symmetry tolerance
    id_offset (int): structure ID starts from id_offset
                        e.g. nstruc = 3, id_offset = 10
                             you obtain ID 10, ID 11, ID 12
    vol_factor (float): volume factor for structure generation
    vol_mu (float): mean for volume scaling
    vol_sigma (float): standard deviation for volume scaling
    vc (bool): variable composition. it needs ll_nat and ul_nat
    ll_nat (tuple): lower limit of number of atoms (e.g. (1, 1))
    ul_nat (tuple): upper limit of number of atoms (e.g. (8, 8))

    # ---------- return
    init_struc_data (dict): {ID: pymatgen Structure, ...}
    '''

    # ---------- initialize
    init_struc_data = {}

    # ---------- Tol_matrix
    tolmat = _set_tol_mat(atype, mindist)

    # ---------- loop for structure generation
    while len(init_struc_data) < nstruc:
        # ------ spgnum --> spg
        if spgnum == 'all':
            spg = random.randint(1, 230)
        else:
            spg = random.choice(spgnum)
        # ------ generate structure
        tmp_crystal = pyxtal()
        if not vc:
            numIons = nat
        else:    # variable composition
            numIons = [random.randint(l, u) for l, u in zip(ll_nat, ul_nat)]
            numIons = tuple(numIons)
        try:
            f = StringIO()    # to get output from pyxtal
            with redirect_stdout(f):
                tmp_crystal.from_random(dim=3, group=spg, species=atype,
                                        numIons=numIons, factor=vol_factor,
                                        conventional=False, tm=tolmat)
            s = f.getvalue().rstrip()    # to delete \n
            if s:
                logger.warning(s)
        except Exception as e:
            logger.warning(f'{e}:     spg = {spg} retry.')
            continue
        if tmp_crystal.valid:
            tmp_struc = tmp_crystal.to_pymatgen(resort=False)
            tmp_nat, _ = get_nat(tmp_struc, atype)
            # -- check the number of atoms
            if tmp_nat != numIons:
                # (pyxtal 0.1.4) cryspy adopts "conventional=False",
                #     which is better for DFT calculation
                # pyxtal returns conventional cell, that is, too many atoms
                tmp_struc = tmp_struc.get_primitive_structure()
                # recheck nat
                if tmp_nat != numIons:    # failure
                    continue
            # -- sort, just in case
            tmp_struc = sort_by_atype(tmp_struc, atype)
            # -- scale volume
            if vol_mu is not None:
                vol = random.gauss(mu=vol_mu, sigma=vol_sigma)
                tmp_struc.scale_lattice(volume=vol)
            # -- check actual space group
            try:
                spg_sym, spg_num = tmp_struc.get_space_group_info(
                    symprec=symprec)
            except TypeError:
                spg_num = 0
                spg_sym = None
            # -- tmp_struc --> init_struc_data
            cid = len(init_struc_data) + id_offset
            init_struc_data[cid] = tmp_struc
            logger.info(f'Structure ID {cid:>6}: {numIons}'
                    f' Space group: {spg:>3} --> {spg_num:>3} {spg_sym}')

    # ---------- return
    return init_struc_data


def gen_struc_mol(
        nstruc,
        atype,
        nat,
        mindist,
        spgnum,
        mol_data,
        nmol,
        symprec=0.01,
        id_offset=0,
        vol_factor=1.1,
        vol_mu=None,
        vol_sigma=None,
        timeout_mol=None,
    ):
    '''
    Generate random molecular crystal structures for given space groups

    # ---------- args
    nstruc (int): number of structures to be generated
    atype (tuple): atom type (e.g. ('Na', 'Cl'))
    nat (tuple): number of atoms (e.g. (4, 4))
    mindist (): minimum interatomic distance (e.g. ((2.0, 1.5), (1.5, 2.0)))
    spgnum (str, int, or tuple): space group number 'all', 0, or tuple of space group numbers
    mol_data (tuple or list): pymatgen Molecule data
    nmol (tuple): number of molecules (e.g. (4, 4))
    symprec (float): symmetry tolerance
    id_offset (int): structure ID starts from id_offset
                        e.g. nstruc = 3, id_offset = 10
                            you obtain ID 10, ID 11, ID 12
    vol_factor (float): volume factor for structure generation
    vol_mu (float): mean for volume scaling
    vol_sigma (float): standard deviation for volume scaling
    timeout_mol (None or float): if float, timeout for molecular structure generation

    # ---------- return
    init_struc_data (dict): {ID: pymatgen Structure, ...}
    struc_mol_id (dict):    not implemented yet, just return vacant dict for now
    '''

    # ---------- initialize
    init_struc_data = {}
    struc_mol_id = {}    # for EA and mol, not implemented yet

    # ---------- Tol_matrix
    tolmat = _set_tol_mat(atype, mindist)

    # ---------- loop for structure generattion
    while len(init_struc_data) < nstruc:
        # ------ spgnum --> spg
        if spgnum == 'all':
            spg = random.randint(1, 230)
        else:
            spg = random.choice(spgnum)
        # ------ generate structure
        if timeout_mol is None:
            tmp_crystal = pyxtal(molecular=True)
            try:
                f = StringIO()    # to get output from pyxtal
                with redirect_stdout(f):
                    tmp_crystal.from_random(
                        dim=3,
                        group=spg,
                        species=mol_data,
                        numIons=nmol,
                        factor=vol_factor,
                        conventional=False,
                        tm=tolmat
                    )
                s = f.getvalue().rstrip()    # to delete \n
                if s:
                    logger.warning(s)
                tmp_valid = tmp_crystal.valid
                if tmp_valid:
                    tmp_struc = tmp_crystal.to_pymatgen(resort=False)
            except Exception as e:
                logger.warning(f'{e}:     spg = {spg} retry.')
                continue
            # if algo in ['EA', 'EA-vc']:
            #     tmp_struc = tmp_crystal.to_pymatgen(resort=False)
            #     tmp_lattice = tmp_struc.lattice
            #     dums = []        # dummy atoms
            #     dum_pos = []     # internal position of dummy
            #     in_dists = []
            #     mol_dists = []
            #     for i, site in enumerate(tmp_crystal.mol_sites):
            #         for j, mol in enumerate(mol_data):
            #             if mol.species == site.mol.species:
            #                 dum = DummySpecie(f"X{j}{i}")
            #                 dums.append(dum)
            #                 dum_pos.append(site.position)
            #                 # -- calculate atom distance by pyxtal
            #                 in_dists.append(np.linalg.norm(site.position.dot(tmp_crystal.lattice.get_matrix()) -
            #                                                 site.get_coords_and_species(absolute=True)[0]
            #                                                 [0:len(site.mol.species)], axis=1))
            #     # -- calculate atom distance from self.mol_data
            #     for j, mol in enumerate(mol_data):
            #         tmp_dists = []
            #         for n, m in enumerate(mol):
            #             tmp_dists.append(mol.get_distance(0, n))
            #         mol_dists.append(tmp_dists)
            #     dum_struc = Structure.from_spacegroup(spg, tmp_lattice, dums, dum_pos)
            #     for ds, dc in zip(dum_struc.species, dum_struc.cart_coords):
            #         tmp_struc.append(ds, dc, coords_are_cartesian=True)
        else:
            # -- multiprocess for hangup prevention
            q = Queue()
            p = Process(target=_mp_mc, args=(tolmat, spg, mol_data, nmol,
                                            vol_factor, q))
            p.start()
            p.join(timeout=timeout_mol)
            if p.is_alive():
                p.terminate()
                p.join()
            if sys.version_info.minor >= 7:
                # Process.close() available from python 3.7
                p.close()
            if q.empty():
                logger.warning('timeout for molecular structure generation. retry.')
                continue
            else:
                # -- get struc data from _mp_mc
                # if rin.algo in ['EA', 'EA-vc']:
                #     tmp_q = q.get()
                #     tmp_struc = tmp_q[0]    # structure data
                #     dums = tmp_q[1]         # dummy element
                #     in_dists = tmp_q[2]     # interatomic distance in a molecule
                #     mol_dists = tmp_q[3]    # distance between molecules
                # else:
                #     tmp_struc = q.get()
                tmp_struc = q.get()
                tmp_valid = q.get()
                if tmp_struc == 'error':
                    # in case of 'error', tmp_valid <-- error message (Exception)
                    logger.warning(tmp_valid.args[0] + f': spg = {spg} retry.')
                    continue
        if tmp_valid:
            # -- scale volume
            if vol_mu is not None:
                vol = random.gauss(mu=vol_mu, sigma=vol_sigma)
                vol = vol * tmp_struc.num_sites / sum(nat)    # for conv. cell
                tmp_struc = scale_cell_mol(tmp_struc, mol_data, vol)
                if not tmp_struc:    # case: scale_cell_mol returns False
                    logger.warning('failed scale cell. retry.')
                    continue
            # -- check nat
            tmp_nat, _ = get_nat(tmp_struc, atype)
            if tmp_nat != nat:
                # cryspy adopts conventional=True
                # pyxtal returns conventional cell,
                # too many atoms if centering
                tmp_struc = tmp_struc.get_primitive_structure()
                # recheck nat
                if tmp_nat != nat:    # failure
                    logger.warning(f'different num. of atoms. {tmp_nat}, {nat} retry.')
                    continue
            # -- grouping atoms for molecule using interatomic distance
            # if rin.algo in ['EA', 'EA-vc']:
            #     loopcnt = 0
            #     while loopcnt < 10:
            #         loopcnt += 1
            #         mol_group = []
            #         group_id = 0
            #         # check atom distance from dummy atom
            #         for i, ts in enumerate(tmp_struc.species):
            #             for dn, dummyatom in enumerate(dums):
            #                 if ts == dummyatom:
            #                     mol_indx = int(ts.symbol[1:2])
            #                     for j, tc_fc in enumerate(tmp_struc.frac_coords):
            #                         if i == j:
            #                             continue
            #                         dist = tmp_struc.get_distance(i, j)
            #                         for dist_indx, true_dist in enumerate(in_dists[dn]):
            #                             if round(true_dist, 2) == round(dist, 2):
            #                                 mol_group.append([tc_fc, group_id, mol_indx, dist_indx])
            #                                 break
            #                     group_id += 1
            #         success_cnt = 0
            #         # check number of atom per group
            #         # if failed, shuffle struc data, and continue(max 10times)
            #         mol_cnt = collections.Counter([x[2] for x in mol_group])
            #         for mc_key, mc_value in mol_cnt.items():
            #             if mc_value % len(mol_data[mc_key]) == 0:
            #                 success_cnt += 1
            #         if success_cnt == len(mol_data):
            #             break
            #         random.shuffle(tmp_struc)
            #     if loopcnt == 10:
            #         continue
            #     tmp_struc.remove_species(dums)
            # -- sort, necessary in molecular crystal
            tmp_struc = sort_by_atype(tmp_struc, atype)
            # -- record group_id, mol_id, and dist_index
            # if rin.algo in ['EA', 'EA-vc']:
            #     tmp_id = []
            #     tmp_mol_indx = []
            #     tmp_dist_indx = []
            #     for atom_fc in tmp_struc.frac_coords:
            #         for gatom in mol_group:
            #             if np.array_equal(atom_fc, gatom[0]):
            #                 tmp_id.append(gatom[1])
            #                 tmp_mol_indx.append(gatom[2])
            #                 tmp_dist_indx.append(gatom[3])
            # -- check minimum distance
            #    from CrySPY 0.10.4
            #    Tol_matrix is used for mindist
            #
            #if self.mindist is not None:
            #    success, mindist_ij, dist = check_distance(tmp_struc,
            #                                               self.atype,
            #                                               self.mindist)
            #    if not success:
            #        print('mindist in gen_struc_mol: {} - {}, {}. retry.'.format(
            #            self.atype[mindist_ij[0]],
            #            self.atype[mindist_ij[1]],
            #            dist), file=sys.stderr)
            #        continue    # failure
            # -- check actual space group
            try:
                spg_sym, spg_num = tmp_struc.get_space_group_info(
                    symprec=symprec)
            except TypeError:
                spg_num = 0
                spg_sym = None
            # -- register the structure in pymatgen format
            cid = len(init_struc_data) + id_offset
            init_struc_data[cid] = tmp_struc
            # if rin.algo in ['EA', 'EA-vc'] and rin.struc_mode in ['mol', 'mol_bs']:
            #     struc_mol_id.update({cid: [tmp_mol_indx, tmp_id, mol_dists]})
            logger.info(f'Structure ID {cid:>6} was generated.'
                    f' Space group: {spg:>3} --> {spg_num:>3} {spg_sym}')

    # ---------- return
    #            struc_mol_id is not implemented yet, just return vacant dict for now
    return init_struc_data, struc_mol_id


def gen_struc_mol_break_sym(
        nstruc,
        atype,
        nat,
        mindist,
        mindist_dummy,
        spgnum,
        mol_data,
        nmol,
        symprec=0.01,
        id_offset=0,
        vol_factor=1.1,
        vol_mu=None,
        vol_sigma=None,
        rot_mol='random_wyckoff',
        nrot=20,
    ):
    '''
    Generate random molecular crystal structures
    molecules are put at a Wyckoff position without consideration of symmetry

    # ---------- args
    nstruc (int): number of structures to be generated
    atype (tuple): atom type (e.g. ('Na', 'Cl'))
    nat (tuple): number of atoms (e.g. (4, 4))
    mindist (): minimum interatomic distance (e.g. ((2.0, 1.5), (1.5, 2.0)))
    mindist_dummy: mindist for dummy atoms (= intermolecular distance)
    spgnum (str, int, or tuple): space group number 'all', 0, or tuple of space group numbers
    mol_data (tuple or list): pymatgen Molecule data
    nmol (tuple): number of molecules (e.g. (4, 4))
    symprec (float): symmetry tolerance
    id_offset (int): structure ID starts from id_offset
                        e.g. nstruc = 3, id_offset = 10
                            you obtain ID 10, ID 11, ID 12
    vol_factor (float): volume factor for structure generation
    vol_mu (float): mean for volume scaling
    vol_sigma (float): standard deviation for volume scaling
    rot_mol (str): rotation option. 'random', 'random_mol', or 'random_wyckoff'
    nrot (int): maximum number of trials to rotate molecules

    # ---------- return
    init_struc_data (dict): {ID: pymatgen Structure, ...}
    struc_mol_id (dict):   not implemented yet, just return vacant dict for now
    '''

    # ---------- initialize
    init_struc_data = {}
    struc_mol_id = {}    # for EA and mol, not implemented yet

    # ------ dummy atom type
    atype_dummy = get_atype_dummy(len(mol_data))

    # ---------- Tol_matrix for dummy atoms
    tolmat = _set_tol_mat(atype_dummy, mindist_dummy)

    # ---------- loop for structure generattion
    while len(init_struc_data) < nstruc:
        # ------ spgnum --> spg
        if spgnum == 'all':
            spg = random.randint(1, 230)
        else:
            spg = random.choice(spgnum)
        # ------ generate structure
        tmp_crystal = pyxtal()
        try:
            f = StringIO()
            with redirect_stdout(f):
                tmp_crystal.from_random(
                    dim=3,
                    group=spg,
                    species=atype_dummy,
                    numIons=nmol,
                    factor=vol_factor,
                    conventional=False,
                    tm=tolmat
                )
            s = f.getvalue().rstrip()    # to delete \n
            if s:
                logger.warning(s)
        except Exception as e:
            logger.warning(f'{e}:    spg = {spg} retry.')
            continue
        if tmp_crystal.valid:
            # -- each wyckoff position --> dummy atom
            dums = []        # dummy atoms
            dum_pos = []     # internal position of dummy
            dum_type = {}    # type of dummy
                                #  e.g. {DummySpecie X00+: 'Rn',
                                #        DummySpecie X10+: 'Rn',
                                #        DummySpecie X20+: 'Xe',
                                #        DummySpecie X30+: 'Xe'}
            for i, site in enumerate(tmp_crystal.atom_sites):
                dum = DummySpecie(f"X{i}")
                dums.append(dum)
                dum_pos.append(site.position)
                dum_type[dum] = site.specie
            # -- tmp_struc with dummy atoms
            lattice = tmp_crystal.lattice.get_matrix()
            tmp_struc = Structure.from_spacegroup(spg, lattice, dums, dum_pos)
            tmp_struc = tmp_struc.get_primitive_structure()
            # -- scale volume
            if vol_mu is not None:
                vol = random.gauss(mu=vol_mu, sigma=vol_sigma)
                tmp_struc.scale_lattice(volume=vol)
            # -- save dummy coords
            dum_species = tmp_struc.species
            dum_coords = tmp_struc.cart_coords
            # if rin.algo in ['EA', 'EA-vc']:
            #     tmp_mol_indx = []
            #     tmp_id = []
            #     tmp_id_cnt = 0
            # -- remove dummy
            tmp_struc.remove_sites(range(0, len(dum_species)))
            tmp_struc_ori = tmp_struc.copy()
            rot_success = False
            # -- rotate molecules
            for nrel in range(nrot):
                tmp_struc = tmp_struc_ori.copy()
                for (dum_specie, dum_coord) in zip(dum_species, dum_coords):
                    mol_index = atype_dummy.index(dum_type[dum_specie])
                    mol = mol_data[mol_index]
                    # rotation option
                    if rot_mol is None:
                        rot_mol_coord = mol.cart_coords
                    if rot_mol == 'random':
                        angle = 2 * np.pi * np.random.rand(3)
                        R = rot_mat(angle)
                        rot_mol_coord = np.matmul(mol.cart_coords, R)
                    if rot_mol == 'random_mol':
                        # each mol_data
                        mol_angles = []    # [angles of mol 1, angles of mol 2, ...]
                        for i in range(len(mol_data)):
                            mol_angles.append(2 * np.pi * np.random.rand(3))
                        angle = mol_angles[mol_index]
                        R = rot_mat(angle)
                        rot_mol_coord = np.matmul(mol.cart_coords, R)
                    if rot_mol == 'random_wyckoff':
                        # each Wyckoff
                        dum_angles = {}    # e.g.
                                            # {DummySpecie X00+: array([ , , ]),
                                            #  DummySpecie X10+: array([ , , ]),
                                            #  DummySpecie X20+: array([ , , ]),
                                            #  DummySpecie X30+: array([ , , ])}
                        angles = 2 * np.pi * np.random.rand(len(dums), 3)
                        for (dum, angle) in zip(dums, angles):
                            dum_angles[dum] = angle
                        angle = dum_angles[dum_specie]
                        R = rot_mat(angle)
                        rot_mol_coord = np.matmul(mol.cart_coords, R)
                    # rotate coord
                    coord = rot_mol_coord + dum_coord
                    # append mol
                    for i, ms in enumerate(mol.species):
                        tmp_struc.append(ms, coord[i], coords_are_cartesian=True)
                        # if rin.algo in ['EA', 'EA-vc']:
                        #     tmp_id.append(tmp_id_cnt)
                        #     tmp_mol_indx.append(mol_index)
                    # if rin.algo in ['EA', 'EA-vc']:
                    #     tmp_id_cnt += 1
                # -- check nat
                tmp_nat, _ = get_nat(tmp_struc, atype)
                if tmp_nat != nat:
                    # pyxtal returns conventional cell,
                    # too many atoms if centering
                    tmp_struc = tmp_struc.get_primitive_structure()
                    # recheck nat
                    if tmp_nat != nat:    # failure
                        if rot_mol is None:
                            break    # go back to the while loop
                        continue
                # -- sort, necessary in molecular crystal
                tmp_struc = sort_by_atype(tmp_struc, atype)
                # -- check minimum distance
                if mindist is not None:
                    success, mindist_ij, dist = check_distance(tmp_struc,
                                                               atype,
                                                               mindist)
                    if not success:
                        type0 = atype[mindist_ij[0]]
                        type1 = atype[mindist_ij[1]]
                        logger.warning(f'mindist: {type0} - {type1}, {dist}. retry.')
                        if rot_mol is None:
                            break    # go back to the while loop
                        continue    # failure
                # -- check actual space group (success)
                try:
                    spg_sym, spg_num = tmp_struc.get_space_group_info(
                        symprec=symprec)
                except TypeError:
                    spg_num = 0
                    spg_sym = None
                rot_success = True
                break
            # -- reach maximum times to rotate (failure)
            if not rot_success:
                continue    # go back to the while loop
            # -- tmp_struc --> init_struc_data
            cid = len(init_struc_data) + id_offset
            init_struc_data[cid] = tmp_struc
            # if rin.algo in ['EA', 'EA-vc'] and rin.struc_mode in ['mol', 'mol_bs']:
            #     # calculate interatomic distance of mol_data
            #     mol_dists = []
            #     for j, mol in enumerate(mol_data):
            #         tmp_dists = []
            #         for n, m in enumerate(mol):
            #             tmp_dists.append(mol.get_distance(0, n))
            #         mol_dists.append(tmp_dists)
            #     struc_mol_id.update({cid: [tmp_mol_indx, tmp_id, mol_dists]})
            logger.info(f'Structure ID {cid:>6} was generated.'
                    f' Space group: {spg:>3} --> {spg_num:>3} {spg_sym}')

    # ---------- return
    #            struc_mol_id is not implemented yet, just return vacant dict for now
    return init_struc_data, struc_mol_id


def _set_tol_mat(atype, mindist):
    tolmat = Tol_matrix()
    for i, itype in enumerate(atype):
        for j, jtype in enumerate(atype):
            if i <= j:
                tolmat.set_tol(itype, jtype, mindist[i][j])
    return tolmat


def _mp_mc(tolmat, spg, mol_data, nmol, vol_factor, q):
    '''
    multiprocess part
    here cannot use rin.xxx and logging
    '''
    try:
        np.random.seed(random.randint(0, 100000000))
        tmp_crystal = pyxtal(molecular=True)
        f = StringIO()
        with redirect_stdout(f):
            with redirect_stderr(f):
                tmp_crystal.from_random(dim=3, group=spg,
                                        species=mol_data, numIons=nmol,
                                        factor=vol_factor, conventional=False, tm=tolmat)
        s = f.getvalue().rstrip()    # to delete \n
        if s:
            logger.warning(s)
        # if algo in ['EA', 'EA-vc']:
        #     tmp_struc = tmp_crystal.to_pymatgen(resort=False)
        #     tmp_lattice = tmp_struc.lattice
        #     dums = []        # dummy atoms
        #     dum_pos = []     # internal position of dummy
        #     in_dists = []
        #     mol_dists = []
        #     for i, site in enumerate(tmp_crystal.mol_sites):
        #         for j, mol in enumerate(mol_data):
        #             if mol.species == site.mol.species:
        #                 dum = DummySpecie(f"X{j}{i}")
        #                 dums.append(dum)
        #                 dum_pos.append(site.position)
        #                 # -- calculate atom distance by pyxtal
        #                 in_dists.append(np.linalg.norm(site.position.dot(tmp_crystal.lattice.get_matrix()) -
        #                                                 site.get_coords_and_species(absolute=True)[0]
        #                                                 [0:len(site.mol.species)], axis=1))
        #     # -- calculate atom distance from self.mol_data
        #     for j, mol in enumerate(mol_data):
        #         tmp_dists = []
        #         for n, m in enumerate(mol):
        #             tmp_dists.append(mol.get_distance(0, n))
        #         mol_dists.append(tmp_dists)
        #     dum_struc = Structure.from_spacegroup(spg, tmp_lattice, dums, dum_pos)
        #     for ds, dc in zip(dum_struc.species, dum_struc.cart_coords):
        #         tmp_struc.append(ds, dc, coords_are_cartesian=True)
        # ---------- queue
        if tmp_crystal.valid:
            # if algo in ['EA', 'EA-vc']:
            #     q.put([tmp_struc, dums, in_dists, mol_dists])
            # else:
            #     q.put(tmp_crystal.to_pymatgen(resort=False))
            q.put(tmp_crystal.to_pymatgen(resort=False))
            q.put(tmp_crystal.valid)
        else:
            q.put(None)
            q.put(tmp_crystal.valid)
    except Exception as e:
        q.put('error')
        q.put(e)



'''
Utility for structures
'''

from logging import getLogger
import os

import numpy as np

from pymatgen.core import Structure, Molecule
from pymatgen.io.cif import CifWriter
from pyxtal.database.collection import Collection
from pyxtal.tolerance import Tol_matrix


logger = getLogger('cryspy')


def set_mindist(atype, mindist_in, factor, struc_mode='crystal',
                dummy=False, mol_file=None, mpi_rank=0):
    # ---------- dummy atom in mol_bs
    if dummy:
        atype = get_atype_dummy(len(mol_file))

    # ---------- mindist
    if mindist_in is None:
        # ------ Tol matrix
        if struc_mode in ['crystal']:
            tolmat = Tol_matrix(prototype='atomic', factor=factor)
        elif struc_mode in ['mol', 'mol_bs']:
            tolmat = Tol_matrix(prototype='molecular', factor=factor)
        else:
            logger.error('struc_mode is wrong')
            raise SystemExit(1)
        # ------ set mindist
        mindist = []
        for i, itype in enumerate(atype):
            tmp = []
            for j, jtype in enumerate(atype):
                tmp.append(tolmat.get_tol(itype, jtype))
            mindist.append(tmp)
    else:
        tmp_array = factor * np.array(mindist_in)
        mindist = tmp_array.tolist()

    # ---------- log
    if mpi_rank == 0:
        for i, itype in enumerate(atype):
            for j, jtype in enumerate(atype):
                if i <= j:
                    if dummy:
                        logger.info(f'{mol_file[i]} - {mol_file[j]}: {mindist[i][j]}')
                    else:
                        logger.info(f'{itype} - {jtype}: {mindist[i][j]}')
        if struc_mode == 'mol':
            logger.info('When struc_mode is mol (only random structure, not EA part),\n'
            '- tolerance between monoatomic molecules is multiplied by 0.8 inside pyxtal (not printed above)\n'
            '- H-N, H-O, or H-F tolerance is multiplied by 0.9 inside pyxtal (not printed above)')

    # ---------- return
    return mindist


def get_atype_dummy(n_mol_file):
    noble_gas = ['Rn', 'Xe', 'Kr', 'Ar', 'Ne', 'He']
    if n_mol_file > len(noble_gas):
        logger.error('len(mol_file) > len(noble_gas)')
        raise SystemExit(1)
    atype = noble_gas[:n_mol_file]
    return atype


def get_mol_data(mol_file):
    '''
    get molecular data

    Input:
        mol_file (tuple)
            e.g. ('Li.xyz', 'PS4.xyz')

    return:
        mol_data (list)
            Molecule data in pymatgen format
    '''
    # ----------
    mol_data = []
    pyxtal_mol_data = Collection('molecules')
    pyxtal_mol_names = list(Collection('molecules'))
    for i, mf in enumerate(mol_file):
        if os.path.isfile(mf):
            mol = Molecule.from_file(mf)
        elif mf in pyxtal_mol_names:
            mol = pyxtal_mol_data[mf]
        else:
            logger.error('no molecular files')
            raise SystemExit(1)
        mol_data.append(mol)
    return mol_data


def out_poscar(struc_data: dict, fpath: str, mode='a'):
    '''
    # ---------- args
    struc_data (dict): {'ID': Structure, 'ID': Structure, ...}
    fpath (str): file path for output
    mode (str): 'w' or 'a'
    '''
    # ---------- if mode='w', clear file
    with open(fpath, mode):
        pass
    # ---------- wrtie POSCAR
    for cid, struc in struc_data.items():
        # ---------- poscar format
        pos = struc.to(fmt='poscar')
        pos = pos.split('\n')
        blank_indx = pos.index('')    # cut unnecessary parts
        pos = pos[:blank_indx]
        pos[0] = f'ID_{cid}'    # replace with ID
        lines = [line+'\n' for line in pos]
        # ---------- always mode='a'
        with open(fpath, 'a') as f:
            for line in lines:
                f.write(line)


def out_cif(struc, cid, tmp_path, fpath, symprec=0.01):
    # ---------- opt_CIFS
    cif = CifWriter(struc, symprec=symprec)
    cif.write_file(tmp_path+'tmp.cif')

    # ---------- correct title for VESTA
    #                (need to delete '_chemical_formula_sum'. i don't know why)
    with open(tmp_path+'tmp.cif', 'r') as fcif:
        ciflines = fcif.readlines()
    ciflines[1] = f'data_ID_{cid}\n'
    if ciflines[11][:21] == '_chemical_formula_sum':
        ciflines.pop(11)
    else:
        logger.error('ciflines[11] is not _chemical_formula_sum,'
                         ' have to fix bag')
        raise SystemExit(1)

    # ---------- cif --> opt_cifs
    with open(fpath, 'a') as foptcif:
        for line in ciflines:
            foptcif.write(line)

    # ---------- clean tmp.cif
    os.remove(tmp_path+'tmp.cif')


def frac_coord_zero_one(struc_in):
    '''
    fractional coordinates: 0.0 <= x,y,z < 1.0
    e.g. [0.0, -0.25, 0.7] --> [0.0, 0.75, 0.7]

    # ---------- args
    struc_in: structure data in pymatgen format
    '''
    struc = struc_in.copy()
    for i in range(struc.num_sites):
        struc[i] = struc[i].to_unit_cell()
    return struc


def origin_shift(struc_in):
    '''
    Randomly shift the origin of struc_in

    # ---------- args
    struc_in: structure data in pymatgen format

    # ---------- return
    origin shifted structure (not change original struc_in)
    '''
    struc = struc_in.copy()
    coords_trans = struc.frac_coords + np.random.rand(3)
    struc_shift = Structure(struc.lattice, struc.species, coords_trans)
    struc_shift = frac_coord_zero_one(struc_shift)
    return struc_shift


def sort_by_atype(struc, atype):
    '''
    return a structre sorted by atype order as a new structure
    '''
    return struc.get_sorted_structure(
        key=lambda x: atype.index(x.species_string))


def check_distance(struc, atype, mindist, check_all=False):
    '''
    # ---------- args
    struc: structure data in pymatgen format
    atype (list): e.g. ['Li', 'Co, 'O']
    mindist (2d list) : e.g. [[2.0, 2.0, 1.2],
                              [2.0, 2.0, 1.2],
                              [1.2, 1.2, 1.5]]
    check_all (bool) : if True, check all atom pairs, return dist_list.
                       if False, stop when (dist < mindist) is found,
                                 return True or False (see below)
    # ---------- return
    (check_all=False) True, None, None: nothing smaller than mindist
    (check_all=False) False, (i, j), dist: something smaller than mindst
                                     here, (i, j) means mindist(i, j)
    (check_all=True) dist_list: if dist_list is vacant,
                                    nothing smaller than mindist
    '''

    # ---------- initialize
    if check_all:
        dist_list = []    # [(i, j, dist), (i, j, dist), ...]

    # ---------- in case there is only one atom
    if struc.num_sites == 1:
        dist = min(struc.lattice.abc)
        if dist < mindist[0][0]:
            if check_all:
                dist_list.append((0, 0, dist))
                return dist_list
            return False, (0, 0), dist
        return True, None, None

    # ---------- normal case
    for i in range(struc.num_sites):
        for j in range(i):
            dist = struc.get_distance(j, i)
            i_type = atype.index(struc[i].species_string)
            j_type = atype.index(struc[j].species_string)
            if dist < mindist[i_type][j_type]:
                if check_all:
                    dist_list.append((j, i, dist))
                else:
                    return False, (i_type, j_type), dist

    # ---------- return
    if check_all:
        if dist_list:
            dist_list.sort()    # sort
            return dist_list
        else:
            return dist_list    # dist_list is vacant list
    return True, None, None


def scale_cell_mol(struc, mol_data, volume, dist_digit=3):
    '''
    scale cell without changing molecule size
    for molecular structures generated by PyXtal
    '''
    # ---------- get molecules
    molecules = []
    for mol in mol_data:
        if len(mol.species) > 1:    # mol.species == 1 <-- atom
            molecules.append(mol)

    # ---------- structure of molecule
    mols_cart_coords = []
    mols_frac_Gs = []
    for mol in molecules:
        # ------ start atom for the molecule
        mol_pos = []
        for i, element in enumerate(struc.species):
            # -- check len
            if len(struc.species) < mol.num_sites + i:
                break
            # -- check number of atoms
            count = 0
            for j in range(mol.num_sites):
                if struc.species[i+j] == mol.species[j]:
                    count += 1
            if count == mol.num_sites:
                mol_pos.append(i)
        # ------ calculate the center of gravity
        in_dists = np.linalg.norm(mol.cart_coords - mol.cart_coords[0], axis=1)
        frac_Gs = []
        mol_cart_coords = []
        # -- take periodic boundary into account
        #    here, check distance from the reference atom (mol.cart_coords[0])
        #    to find a correct site for calculating the center of gravity
        for st in mol_pos:
            frac_mol_coords = struc.frac_coords[st:st + mol.num_sites]
            cart_mol_basis = struc.lattice.get_cartesian_coords(frac_mol_coords[0])
            fix_frac_coords = []
            for frac, dist in zip(frac_mol_coords, in_dists):
                all_pattern = [[frac[0]], [frac[1]], [frac[2]]]
                basis_dist = frac - frac_mol_coords[0]
                for axis in range(3):
                    if basis_dist[axis] > 0:
                        all_pattern[axis].append(frac[axis]-1)
                    elif basis_dist[axis] < 0:
                        all_pattern[axis].append(frac[axis]+1)
                # check all patterns
                break_loop = False
                for a_coord in all_pattern[0]:
                    for b_coord in all_pattern[1]:
                        for c_coord in all_pattern[2]:
                            tmp_frac_coord = [a_coord, b_coord, c_coord]
                            tmp_cart_coord = struc.lattice.get_cartesian_coords(tmp_frac_coord)
                            tmp_cart_dist = np.linalg.norm(tmp_cart_coord - cart_mol_basis)
                            if round(dist, dist_digit) == round(tmp_cart_dist, dist_digit):
                                break_loop = True
                                fix_frac_coords.append(tmp_frac_coord)
                                break
                        if break_loop:
                            break
                    if break_loop:
                        break
            if not len(fix_frac_coords) == mol.num_sites:
                return False
            fix_frac_coords = np.array(fix_frac_coords)
            # -- get the center of gravity
            frac_G = fix_frac_coords.sum(axis=0)/mol.num_sites
            frac_Gs.append(frac_G)
            # -- save structure of mol
            fix_cart_coords = struc.lattice.get_cartesian_coords(fix_frac_coords)
            cart_G = fix_cart_coords.sum(axis=0)/mol.num_sites
            mol_cart_coord = fix_cart_coords - cart_G
            mol_cart_coords.append(mol_cart_coord)
        mols_frac_Gs.append(frac_Gs)
        mols_cart_coords.append(mol_cart_coords)
        # ------ remove the molecule from sturc
        for st in reversed(mol_pos):
            struc.remove_sites(range(st, st + mol.num_sites))

    # ---------- scale lattice
    struc.scale_lattice(volume)

    # ---------- put the molecule of original size
    for mol_index in range(len((molecules))):
        molecule = molecules[mol_index]
        for frac_G, cart_mol in zip(mols_frac_Gs[mol_index],
                                    mols_cart_coords[mol_index]):
            cart_G = struc.lattice.get_cartesian_coords(frac_G)
            cart_mol_coord = cart_mol + cart_G
            for spe, spe_coord in zip(molecule.species, cart_mol_coord):
                struc.append(spe, spe_coord, coords_are_cartesian=True)

    # ---------- return new struc
    return struc


def rot_mat(angles, seq='zyx', degree=False):
    '''
    Calculate rotation matrix

    # ---------- input
    angles: in radian (degree=False)
    seq:
    degree:
    '''
    if not len(seq) == len(angles):
        logger.error('not len(seq) == len(angles)')
        raise SystemExit(1)

    if degree:
        angles = np.deg2rad(angles)

    R = np.eye(3)
    angles = angles[::-1]
    seq = seq[::-1]
    for axis, angle in zip(seq, angles):
        if 'x' == axis:
            tmp_R = np.array([[1, 0, 0],
                              [0, np.cos(angle), -np.sin(angle)],
                              [0, np.sin(angle), np.cos(angle)]])
        elif 'y' == axis:
            tmp_R = np.array([[np.cos(angle), 0, np.sin(angle)],
                              [0, 1, 0],
                              [-np.sin(angle), 0, np.cos(angle)]])
        elif 'z' == axis:
            tmp_R = np.array([[np.cos(angle), -np.sin(angle), 0],
                              [np.sin(angle), np.cos(angle), 0],
                              [0, 0, 1]])
        else:
            tmp_R = np.eye(3)
        R = np.matmul(R, tmp_R)

    return R


def find_site(struc, mol_id, group_id, true_dists):
    '''
    find out atom on the unit cell, because don't refer atom in neighbor cell
    '''
    # ---------- init
    fix_frac_coords = []
    fix_group_id = []
    fix_mol_id = []
    grouped_struc_mol_id = []
    id_cnt = 0
    grouped_struc = []
    # ----------  separate molecules per group_id
    for j in range(len(struc)):
        remove_indx = []
        for i, s in enumerate(struc):
            if id_cnt != group_id[i]:
                remove_indx.append(i)
        if len(remove_indx) != len(struc):
            tmp_struc = struc.copy()
            tmp_struc.remove_sites(remove_indx)
            grouped_struc.append(tmp_struc)
            for n in range(len(struc)):
                if group_id[n] == id_cnt:
                    grouped_struc_mol_id.append(mol_id[n])
                    break
        id_cnt += 1
    fix_species = []
    # ---------- compare atom distance to atom distance of mol_data
    for j, ts in enumerate(grouped_struc):
        for i, frac in enumerate(ts.frac_coords):
            dist = true_dists[grouped_struc_mol_id[j]][i]
            all_pattern = [[frac[0]], [frac[1]], [frac[2]]]
            basis_dist = frac - ts[0].frac_coords
            for axis in range(3):
                if basis_dist[axis] >= 0:
                    all_pattern[axis].append(frac[axis]-1)
                elif basis_dist[axis] < 0:
                    all_pattern[axis].append(frac[axis]+1)
            # ------ check all patterns
            # break_loop = False
            min_dist = 1000000
            for a_coord in all_pattern[0]:
                for b_coord in all_pattern[1]:
                    for c_coord in all_pattern[2]:
                        tmp_frac_coord = [a_coord, b_coord, c_coord]
                        tmp_cart_coord = struc.lattice.get_cartesian_coords(tmp_frac_coord)
                        base_cart = struc.lattice.get_cartesian_coords(ts.frac_coords[0])
                        tmp_cart_dist = np.linalg.norm(tmp_cart_coord - base_cart)
                        dist_rmse = np.sqrt(np.mean((dist - tmp_cart_dist)**2))
                        if dist_rmse < min_dist:
                            min_dist = dist_rmse
                            min_dist_frac_coord = tmp_frac_coord
            fix_frac_coords.append(min_dist_frac_coord)
            fix_group_id.append(j)
            for n, gid in enumerate(group_id):
                if gid == j:
                    fix_mol_id.append(grouped_struc_mol_id[j])
                    break
            fix_species.append(grouped_struc[j][i].specie)
    return np.array(fix_frac_coords), fix_group_id, fix_species, fix_mol_id


def cal_g(struc, mol_id, group_id, true_dists):
    # ----------  calculate molecule center
    mol_g_A = []
    frac_coords, fix_group_id, _, _ = find_site(struc, mol_id, group_id, true_dists)
    for cnt in range(len(struc)):
        acnt = 0
        cal_g = 0
        for i, fc in enumerate(frac_coords):
            if cnt == fix_group_id[i]:
                acnt += 1
                cal_g += fc
        if acnt == 0:
            break
        mol_g_A.append(cal_g/acnt)
    return mol_g_A


def sort_by_atype_mol(struc, atype, mol_id, group_id):
    # ---------- work the same as sort_by_atype, and sort mol_id and group_id
    copy_struc = struc.copy()
    sorted_struc = sort_by_atype(struc, atype)
    sorted_mol_id = []
    sorted_group_id = []
    for i in range(len(sorted_struc)):
        for j in range(len(copy_struc)):
            if np.allclose(sorted_struc.frac_coords[i], copy_struc.frac_coords[j]):
                sorted_mol_id.append(mol_id[j])
                sorted_group_id.append(group_id[j])
    return sorted_struc, sorted_mol_id, sorted_group_id


def get_nat(struc, atype):
    compos = struc.composition.as_dict()
    nat = tuple([int(compos[at]) for at in atype])
    ratio = tuple([x/sum(nat) for x in nat])
    return nat, ratio


from logging import getLogger

from ..IO import io_stat, pkl_data


logger = getLogger('cryspy')

def initialize(rin):
    logger.info('# ---------- Initialize LAQA')

    # ---------- initialize
    tot_step_select = [0]
    laqa_step = {}
    laqa_struc = {}
    laqa_energy = {}
    laqa_bias = {}
    laqa_score = {}
    for i in range(rin.tot_struc):
        laqa_step[i] = []
        laqa_struc[i] = []
        laqa_energy[i] = []
        laqa_bias[i] = []
        laqa_score[i] = [float('inf')]
    id_queueing = [i for i in range(rin.tot_struc)]
    id_select_hist = []
    id_running = []

    # ---------- save for LAQA
    pkl_data.save_id_queueing(id_queueing)
    pkl_data.save_id_running(id_running)
    pkl_data.save_id_select_hist(id_select_hist)
    pkl_data.save_tot_step_select(tot_step_select)
    pkl_data.save_laqa_step(laqa_step)
    pkl_data.save_laqa_struc(laqa_struc)
    pkl_data.save_laqa_energy(laqa_energy)
    pkl_data.save_laqa_bias(laqa_bias)
    pkl_data.save_laqa_score(laqa_score)

    # ---------- status
    stat = io_stat.stat_read()
    io_stat.set_common(stat, 'selection', 0)
    io_stat.set_common(stat, 'total_step', 0)
    io_stat.set_id(stat, 'selected_id', id_queueing)    # all IDs
    io_stat.set_id(stat, 'id_queueing', id_queueing)    # all IDs
    io_stat.write_stat(stat)

    # ---------- log
    logger.info('# ---------- Selection 0')
    if len(id_queueing) > 30:
        logger.info(f'selected_id: {len(id_queueing)} IDs')
    else:
        x = ' '.join(str(a) for a in id_queueing)
        logger.info(f'selected_id: {x}')



from logging import getLogger
import numpy as np

from . import structure as lammps_structure


logger = getLogger('cryspy')


def collect_lammps(rin, cid, work_path, nat):
    # ---------- natot
    natot = sum(nat)    # do not use rin.natot here for EA-vc
    # ---------- check optimization in current stage & obtain energy
    energy = np.nan
    check_opt = 'not_yet'
    try:
        with open(work_path+rin.lammps_outfile, 'r') as fout:
            lines = fout.readlines()
        for i, line in enumerate(lines):
            if 'ERROR:' in line:
                logger.warning('    ' + line.rstrip())
                energy = np.nan
                check_opt = 'not_yet'
                break
            elif 'Minimization stats:' in line:
                energy = float(lines[i+3].split()[2])  # in eV (units is metal)
                energy = energy/float(natot)    # eV/cell --> eV/atom
                check_opt = 'done'
    except Exception as e:
        logger.warning(f'{e}:    Structure ID {cid},'
                       f'could not obtain energy from {rin.lammps_outfile}')
        energy = np.nan    # error
        check_opt = 'no_file'

    # ---------- obtain magmom
    magmom = np.nan    # magnetic moment is not calculated

    # ---------- collect the last structure
    try:
        opt_struc = lammps_structure.from_file(rin, work_path+'log.struc', nat)
    except Exception as e:
        logger.warning(e)
        opt_struc = None

    # ---------- return
    return opt_struc, energy, magmom, check_opt



from logging import getLogger

import numpy as np
from pymatgen.analysis.structure_matcher import StructureMatcher


logger = getLogger('cryspy')


def survival_fittest(fitness, struc_data, elite_struc=None, elite_fitness=None,
                     n_fittest=0, fit_reverse=False, emax_ea=None, emin_ea=None):
    '''
    # ---------- args
    fitness (dict): {ID: fitness, ...}
    struc_data (dict): {ID: structure data, ...}
    elite_struc (dict): {ID: elite structure data, ...}
    elite_fitness (dict): {ID: elite fitness, ...}
    n_fittest (int): number of fittest data which can survive
    fit_reverse (bool): if False, lower fitness is better
    emax_ea (float): maximum energy for cutoff
    emin_ea (float): minimum energy for cutoff

    # ---------- return
    ranking (list): [ID, ...]
    fit_with_elite (dict): {ID: fitness, ...} fitness + elite_fitness
    '''

    # ---------- initialize
    struc_with_elite = struc_data.copy()    # shallow copy to leave original data
    fit_with_elite = fitness.copy()          # shallow copy to leave original data
    ncheck = 5
    ranking = []
    smatcher = StructureMatcher()    # instantiate

    # ---------- elite
    if elite_struc is not None:
        struc_with_elite.update(elite_struc)
        fit_with_elite.update(elite_fitness)

    # ---------- None, np.nan, emax_ea, emin_ea
    for cid, value in fit_with_elite.items():
        if value is None or np.isnan(value):
            fit_with_elite[cid] = -np.inf if fit_reverse else np.inf
        # ------ emax_ea
        if emax_ea is not None:
            if value > emax_ea:
                fit_with_elite[cid] = -np.inf if fit_reverse else np.inf
                logger.info(f'Eliminate ID {cid}: {value} > emax_ea')
        # ------ emin_ea
        if emin_ea is not None:
            if value < emin_ea:
                fit_with_elite[cid] = -np.inf if fit_reverse else np.inf
                logger.info(f'Eliminate ID {cid}: {value} < emin_ea')

    # ---------- sort fit_with_elite
    sorted_fit_with_elite = sorted(fit_with_elite, key=fit_with_elite.get, reverse=fit_reverse)

    # ---------- ranking without duplication
    for cid in sorted_fit_with_elite:
        # ------ init dupl_flag
        dupl_flag = False
        # ------ for structure is None
        if struc_with_elite[cid] is None:
            continue    # next cid
        # ------ duplication check
        for tid in ranking[:-(ncheck+1):-1]:
            if smatcher.fit(struc_with_elite[cid], struc_with_elite[tid]):
                dupl_flag = True
                break
        # ------ register or skip
        if dupl_flag:
            continue    # next cid
        else:
            ranking.append(cid)
            # -- n_fittest
            if n_fittest > 0:
                if len(ranking) == n_fittest:
                    break

    # ---------- return
    return ranking, fit_with_elite, struc_with_elite


'''
Collect results in OpenMX
written  by  H. Sawahata 2020/03/09
info at hikaruri.jp
'''

from logging import getLogger

import numpy as np
import re
from pymatgen.core.units import Energy

from . import structure as OMX_structure


logger = getLogger('cryspy')

def collect_OMX(rin, cid, work_path, nat):
    # ---------- check optimization in previous stage (done)
    # If *.out file exists, the calculation is done.
    try:
        with open(work_path+rin.OMX_outfile, 'r') as fpout:
            lines = fpout.readlines()
        check_opt = 'done'
    except Exception as e:
        logger.warning(e)
        check_opt = 'no_file'

    # ---------- obtain energy and magmom (done)
    natot = sum(nat)    # do not use rin.natot here for EA-vc
    try:
        with open(work_path+rin.OMX_outfile, 'r') as fpout:
            lines = fpout.readlines()
        energy = np.nan
        for line in reversed(lines):
            if re.search('Utot\.', line):
                energy = float(re.search(r"-\d.+", line).group())
                energy = float(Energy(energy, 'Ha').to('eV'))
                energy = energy / float(natot)
                break
        magmom = np.nan # implemented (2020/10/03)
        for line in lines:
            if line.find("muB") >= 0:
                muB = line.split()
                magmom = float(muB[4])
                break
    except Exception as e:
        energy = np.nan    # error
        magmom = np.nan    # error
        logger.warning(f'{e}:    Structure ID {cid},'
                       f' could not obtain energy from {rin.OMX_outfile}')

    # ---------- collect the last structure (yet)
    try:
        lines_cell = OMX_structure.extract_cell_parameters_from_outfile(
            work_path+rin.OMX_outfile)
        if lines_cell is None:
            lines_cell = OMX_structure.extract_cell_parameters_from_infile(
            work_path+rin.OMX_infile)
        lines_atom = OMX_structure.extract_atomic_positions_from_outfile(
            work_path+rin.OMX_outfile, nat)
        if lines_atom is None:
            lines_atom = OMX_structure.extract_atomic_positions_from_infile(
                rin, work_path+rin.OMX_infile, nat)
        opt_struc = OMX_structure.from_lines(lines_cell, lines_atom)
        # ------ opt_OMX-structure
        with open('./data/opt_OMX-structure', 'a') as fstruc:
            fstruc.write(f'# ID {cid}\n')
        OMX_structure.write(rin, opt_struc, './data/opt_OMX-structure', mode='a')
    except Exception as e:
        logger.warning(e)
        opt_struc = None

    # ---------- check
    if np.isnan(energy):
        opt_struc = None
    if opt_struc is None:
        energy = np.nan
        magmom = np.nan

    # ---------- return
    return opt_struc, energy, magmom, check_opt



from logging import getLogger
import os


logger = getLogger('cryspy')


def check_input_vasp(rin):
    # ---------- prepare rin.jobfile, POTCAR, INCAR
    calc_inputs = [rin.jobfile, 'POTCAR', 'INCAR']

    # ------ check required files
    for f in calc_inputs:
        if f == 'INCAR':
            fincars = [f'INCAR_{i}' for i in range(1, rin.nstage+1)]
            for ff in fincars:
                if not os.path.isfile('./calc_in/' + ff):
                    logger.error('Could not find ./calc_in/' + ff)
                    raise SystemExit(1)
        else:
            if not os.path.isfile('./calc_in/' + f):
                logger.error('Could not find ./calc_in/' + f)
                raise SystemExit(1)



from logging import getLogger
import os

from ...IO import pkl_data


logger = getLogger('cryspy')


def collect_ext():
    # ---------- get opt_struc_data and energy_data
    if os.path.isdir('ext/calc_data'):
        ext_opt_struc_data = pkl_data.load_ext_opt_struc()
        ext_energy_data = pkl_data.load_ext_energy()
    else:
        logger.error('no ext/calc_data directory')
        raise SystemExit(1)
    # ---------- return
    return ext_opt_struc_data, ext_energy_data



from logging import getLogger
import os

from ..start import cryspy_restart

from ..IO import io_stat, pkl_data
from ..job.ctrl_job import prepare_jobfile, submit_next_struc, regist_opt, mv_fin, next_gen_EA
from ..util.utility import get_version
#from ..start.gen_init_struc import gen_init_struc

# ---------- import later (after rin = ReadInput())
#from ..interface import select_code


logger = getLogger('cryspy')


def restart_interact(njob: int):
    # ---------- logger etc
    comm = None
    mpi_rank = 0
    mpi_size = 1

    # ---------- restart
    rin, init_struc_data = cryspy_restart.restart(comm, mpi_rank, mpi_size)
    if rin.nstage > 1:
        logger.warning('nstage is ignored in interactive mode')
    if njob == 0:
        njob = rin.njob

    # ---------- check calc files in ./calc_in
    from ..interface import select_code    # after rin = ReadInput()
    select_code.check_calc_files(rin)

    # ---------- mkdir work/fin
    os.makedirs('work/fin', exist_ok=True)

    # ----------
    # ---------- Ctrl_job
    # ----------

    # ---------- load data
    opt_struc_data = pkl_data.load_opt_struc()
    rslt_data = pkl_data.load_rslt()
    id_queueing = pkl_data.load_id_queueing()
    id_running = pkl_data.load_id_running()
    if rin.algo == 'RS':
        gen = None
    elif rin.algo == 'EA':
        gen = pkl_data.load_gen()

    # ---------- flag for next selection or generation
    if rin.algo in ['BO', 'LAQA', 'EA', 'EA-vc']:
        if (id_queueing or id_running):
            go_next_sg = False
        else:
            go_next_sg = True

    # ----------  check_job
    if not rin.stop_next_struc:
        while len(id_running) < njob and id_queueing:
            id_running.append(id_queueing.pop(0))

    # in interactive mode, tmp_id_running is not used
    # because id_running does not change during the for loop below
    for cid in id_running:
        # --------- mkdir
        os.makedirs(f'work/{cid}', exist_ok=True)
        work_path = f'work/{cid}/'

        # ---------- next_struc
        #            interactive mode always starts with next_struc
        select_code.next_struc(rin, init_struc_data[cid], cid, work_path, rin.nat)

        # ---------- prepare jobfile
        prepare_jobfile(rin, cid, work_path)

        # ---------- submit job
        submit_next_struc(rin, cid, work_path, wait=True)

        # ---------- collect
        opt_struc, energy, magmom, check_opt = \
            select_code.collect(rin, cid, work_path, rin.nat)
        logger.info(f'{cid}    collect results: E = {energy} eV/atom')

        # ---------- register opt data
        opt_struc_data, rslt_data = regist_opt(
            rin, cid, work_path,
            init_struc_data, opt_struc_data, rslt_data,
            opt_struc, energy, magmom, check_opt, ef=None, n_selection=None, gen=gen)

        # ---------- mv work to fin
        mv_fin(cid)

    # ---------- id
    id_running.clear()
    pkl_data.save_id_queueing(id_queueing)
    pkl_data.save_id_running(id_running)
    stat = io_stat.stat_read()
    io_stat.set_id(stat, 'id_queueing', id_queueing)
    io_stat.write_stat(stat)

    # ---------- next selection or generation
    if not (id_queueing or id_running):
        # ---------- RS
        if rin.algo == 'RS':
            logger.info('\nDone all structures!')
        # ---------- EA
        if rin.algo == 'EA':
            next_gen_EA(
                rin,
                id_queueing,
                id_running,
                gen,
                go_next_sg,
                init_struc_data,
                opt_struc_data,
                rslt_data,
                ea_vc_data=None,
                struc_mol_id=None,
            )


'''
Structure file for LAMMPS
'''
import itertools

import numpy as np
from pymatgen.core import Structure


def from_file(rin, name, nat):
    # ---------- natot
    natot = sum(nat)    # do not use rin.natot here for EA-vc
    # ---------- last structure
    with open(name, 'r') as f:
        lines = f.readlines()
    for i, line in enumerate(lines):
        if 'ITEM: TIMESTEP' in line:
            tmp_lines = lines[i:i+natot+9]
    lines = tmp_lines

    # ---------- lattice
    xlo_bound, xhi_bound, xy = [float(x) for x in lines[5].split()]
    ylo_bound, yhi_bound, xz = [float(x) for x in lines[6].split()]
    zlo_bound, zhi_bound, yz = [float(x) for x in lines[7].split()]

    xlo = xlo_bound - min(0.0, xy, xz, xy+xz)
    xhi = xhi_bound - max(0.0, xy, xz, xy+xz)
    ylo = ylo_bound - min(0.0, yz)
    yhi = yhi_bound - max(0.0, yz)
    zlo = zlo_bound
    zhi = zhi_bound

    a = [xhi-xlo, 0, 0]
    b = [xy, yhi-ylo, 0]
    c = [xz, yz, zhi-zlo]

    lattice = np.array([a, b, c])    # in Ang, each row is a lattice vector

    # ---------- non replica coordinates
    coords = [[float(x) for x in line.split()][2:] for line in lines[9:]]

    # ---------- species
    species = [itertools.repeat(typ, times=num) for typ, num in zip(
        rin.atype, nat)]
    species = list(itertools.chain.from_iterable(species))

    structure = Structure(lattice, species, coords)

    return structure


def write(rin, structure, output, nat, title="LAMMPS"):
    # ---------- convert lattice constants and angles to LAMMPS format
    cos = lambda x: np.cos(np.deg2rad(x))
    sqrt = lambda x: np.sqrt(x)

    a, b, c = structure.lattice.abc
    alpha, beta, gamma = structure.lattice.angles

    lx = a
    xy = b*cos(gamma)
    xz = c*cos(beta)
    ly = sqrt(b**2 - xy**2)
    yz = (b*c*cos(alpha) - xy*xz) / ly
    lz = sqrt(c**2 - xz**2 - yz**2)

    # ---------- coordinates
    natot = sum(nat)
    index = range(1, natot+1)
    species = [itertools.repeat(typ, times=num) for typ, num in zip(
        index, nat)]
    species = list(itertools.chain.from_iterable(species))
    sites = structure.sites

    # ---------- write
    with open(output, 'w') as f:
        f.write('# data_{0:s}\n'.format(''.join(title.split())))
        f.write('\n')
        f.write(str(natot) + ' atoms\n')
        f.write(str(len(rin.atype)) + ' atom types\n')
        f.write('\n')
        f.write('0.000000  {0:10.6f}   xlo xhi\n'.format(lx))
        f.write('0.000000  {0:10.6f}   ylo yhi\n'.format(ly))
        f.write('0.000000  {0:10.6f}   zlo zhi\n'.format(lz))
        f.write('\n')
        f.write('{0:10.6f}  {1:10.6f}  {2:10.6f}   xy xz yz\n'.format(
            xy, xz, yz))
        f.write('\n')
        f.write('Atoms\n')
        f.write('\n')

        for i, k, site in zip(index, species, sites):
            f.write('{0:4d}  {1:<4d}   {2[0]:7f} {2[1]:7f} {2[2]:7f}\n'.format(
                i, k, site.coords))



from logging import getLogger
import os

from ..start import cryspy_init
from ..util.utility import set_logger, backup_cryspy, clean_cryspy

from .restart_interact import restart_interact
from .rslt_energy import display_rslt
from .view_atom import get_ase_atoms
from .energy_plot import energy_plot


set_logger()
logger = getLogger('cryspy')


def initialize():
    comm = None
    mpi_rank = 0
    mpi_size = 1
    if os.path.isfile('lock_cryspy'):
        logger.error('lock_cryspy file exists')
        raise SystemExit(1)
    with open('lock_cryspy', 'w'):
        pass    # create vacant file
    if not os.path.isfile('cryspy.stat'):
        cryspy_init.initialize(comm, mpi_rank, mpi_size)
    else:
        logger.error('cryspy.stat file exists. Clean files to start from the beginning.')
    os.remove('lock_cryspy')

def backup():
    backup_cryspy()


def clean():
    clean_cryspy()


def restart(njob: int):
    if os.path.isfile('lock_cryspy'):
        logger.error('lock_cryspy file exists')
        raise SystemExit(1)
    with open('lock_cryspy', 'w'):
        pass    # create vacant file
    if os.path.isfile('cryspy.stat'):
        restart_interact(njob)
    else:
        logger.error('cryspy.stat file does not exist.')
    os.remove('lock_cryspy')


def show_rslt(cid='all'):
    display_rslt(cid)


def get_atoms(status, cid):
    return get_ase_atoms(status, cid)


def plot_e(y_max=2.0, y_min=-0.5):
    energy_plot(y_max, y_min)


from logging import getLogger
import os
import shutil

from . import structure as lammps_structure


logger = getLogger('cryspy')


def next_stage_lammps(rin, stage, work_path, nat):
    # ---------- skip_flag
    skip_flag = False

    # ---------- rename lammps files at the current stage
    lammps_files = [rin.lammps_infile, rin.lammps_outfile,
                    rin.lammps_data, 'log.struc']
    for file in lammps_files:
        if not os.path.isfile(work_path + file):
            logger.error('Not found ' + work_path + file)
            raise SystemExit(1)
        os.rename(work_path + file, work_path + f'stage{stage}_' + file)

    # ---------- copy the input file from ./calc_in for the next stage
    finfile = './calc_in/' + rin.lammps_infile + f'_{stage + 1}'
    shutil.copyfile(finfile, work_path+rin.lammps_infile)

    # ---------- generate the structure data file
    try:
        structure = lammps_structure.from_file(
            rin, work_path + f'stage{stage}_log.struc', nat)
    except ValueError:
        skip_flag = True
        logger.warning('    error in lammps,  skip this structure')
        return skip_flag
    with open(work_path + f'stage{stage}_' + rin.lammps_data, 'r') as f:
        lines = f.readlines()
    title = lines[0][7:]    # string following 'data_'
    lammps_structure.write(rin,
                           structure,
                           work_path+rin.lammps_data,
                           nat,
                           title=title)

    # ---------- return
    return skip_flag


def next_struc_lammps(rin, structure, cid, work_path, nat):
    # ---------- copy files
    if rin.lammps_potential is None:
        calc_inputs = [rin.lammps_infile]
    else:
        calc_inputs = [rin.lammps_infile] + rin.lammps_potential
    for f in calc_inputs:
        ff = f+'_1' if f == rin.lammps_infile else f
        if not os.path.isfile('./calc_in/'+ff):
            logger.error('Could not find ./calc_in/'+ff)
            raise SystemExit(1)
        shutil.copyfile('./calc_in/'+ff, work_path+f)

    # ---------- generate the structure data file
    lammps_structure.write(rin,
                           structure,
                           work_path+rin.lammps_data, nat,
                           title=f'ID_{cid}')



'''
Restart CrySPY
'''

from datetime import datetime
from logging import getLogger
import os

from ..IO import diff_input, pkl_data
from ..IO.read_input import ReadInput
from ..RS.rs_gen import gen_random
from ..util.utility import get_version, backup_cryspy
from ..util.struc_util import out_poscar

# ---------- import later
#from ..RS import rs_restart
#from ..BO import bo_restart
#from ..LAQA import laqa_restart
#from ..EA import ea_append


logger = getLogger('cryspy')

def restart(comm, mpi_rank, mpi_size):
    if mpi_rank == 0:
        logger.info('\n\n\nRestart CrySPY ' + get_version() + '\n\n')

    # ########## MPI start
    if mpi_size > 1:
        comm.barrier()
    # ---------- read input and check the change
    if mpi_rank == 0:
        logger.info('read input, cryspy.in')
    try:
        # all processes read input data
        rin = ReadInput()    # read input data, cryspy,in
    except Exception as e:
        if mpi_rank == 0:
            logger.error(e)
        raise SystemExit(1)
    if mpi_rank == 0:
        try:
            # only rank0 compares current and previous input
            pin = pkl_data.load_input()    # load previous input data from input_data.pkl
            diff_input.diff_in(rin, pin)           # compare current and previous input
            pkl_data.save_input(rin)       # save input data to input_data.pkl
        except Exception as e:
            if mpi_size > 1:
                comm.Abort(1)
            logger.error(e)
            raise SystemExit(1)

    # ------ load init_struc_data for appending structures
    # In EA, one can not change tot_struc, so struc_mol_id need not be considered here
    # _append_struc is not allowed in EA and EA-vc either
    init_struc_data = pkl_data.load_init_struc()
    append_flag = False

    # ---------- append structures
    if len(init_struc_data) < rin.tot_struc:
        # ------ append_flag
        append_flag = True
        # ------ backup
        if mpi_rank == 0:
            backup_cryspy()
        # ------ barrier for MPI
        if mpi_size > 1:
            comm.barrier()
        # ------ append struc.
        if mpi_rank == 0:
            prev_nstruc = len(init_struc_data)
        #        init_struc_data is saved in _append_struc
        init_struc_data = _append_struc(rin, init_struc_data, comm, mpi_rank, mpi_size)
    elif rin.tot_struc < len(init_struc_data):
        logger.error('tot_struc < len(init_struc_data)')
        raise SystemExit(1)

    # ---------- append structures by EA (option)
    # not support MPI
    if rin.append_struc_ea:
        # ------ append_flag
        append_flag = True
        if mpi_rank == 0:
            # ------ backup
            backup_cryspy()
    #
    # struc_mol_id has not developed yet here
    #if rin.struc_mode in ['mol', 'mol_bs']:
    #    struc_mol_id = pkl_data.load_struc_mol_id()
    #
            from ..EA import ea_append
            prev_nstruc = len(init_struc_data)
            # init_struc_data is saved in ea_append.append_struc()
            init_struc_data = ea_append.append_struc(rin, init_struc_data)

    # ---------- post append
    if append_flag:
        if mpi_rank == 0:
            # ------ RS
            if rin.algo == 'RS':
                from ..RS import rs_restart
                rs_restart.restart(rin, prev_nstruc)
            # ------ BO
            if rin.algo == 'BO':
                from ..BO import bo_restart
                bo_restart.restart(rin, init_struc_data, prev_nstruc)
            # ------ LAQA
            if rin.algo == 'LAQA':
                from ..LAQA import laqa_restart
                laqa_restart.restart(rin, prev_nstruc)
            # ------ remove lock_cryspy
            os.remove('lock_cryspy')
        raise SystemExit()

    # ---------- return
    return rin, init_struc_data


def _append_struc(rin, init_struc_data, comm, mpi_rank, mpi_size):
    # ---------- log
    if mpi_rank == 0:
        logger.info('# ---------- Append structures')
    if mpi_size > 1:
        comm.barrier()

    # ---------- time
    if mpi_rank == 0:
        time_start = datetime.today()

    # ---------- generate structures
    # only init_struc_data in rank0 is important
    nstruc = rin.tot_struc - len(init_struc_data)
    tmp_struc_data, tmp_struc_mol_id = gen_random(rin, nstruc, len(init_struc_data),
                                        comm, mpi_rank, mpi_size)

    if mpi_rank == 0:
        out_poscar(tmp_struc_data, './data/init_POSCARS')
        # ---------- update and save
        init_struc_data.update(tmp_struc_data)
        # ToDo: struc_mol_id
        pkl_data.save_init_struc(init_struc_data)

        # ---------- time
        time_end = datetime.today()
        etime = time_end - time_start
        logger.info(f'Elapsed time for structure generation: {etime}')

    # ---------- return
    # only init_struc_data in rank0 is important
    return init_struc_data


from ..IO import io_stat, pkl_data


def initialize(rin):
    # ---------- initialize
    id_queueing = [i for i in range(rin.tot_struc)]
    id_running = []

    # ---------- status
    stat = io_stat.stat_read()
    io_stat.set_id(stat, 'id_queueing', id_queueing)
    io_stat.write_stat(stat)

    # ---------- save
    pkl_data.save_id_queueing(id_queueing)
    pkl_data.save_id_running(id_running)



'''
Structure file for Quantum ESPRESSO
'''

from logging import getLogger

from pymatgen.core import Structure
from pymatgen.core.units import Length


logger = getLogger('cryspy')

def extract_cell_parameters(filename):
    # ---------- last CELL_PARAMETERS
    with open(filename, 'r') as f:
        lines = f.readlines()
    lines_cell = None
    for i, line in enumerate(reversed(lines)):
        if 'CELL_PARAMETERS' in line:
            ibegin = len(lines) - 1 - i
            iend = ibegin + 4
            lines_cell = lines[ibegin:iend]
            break
    return lines_cell


def extract_atomic_positions(filename, nat):
    # ---------- last ATOMIC_POSITIONS
    with open(filename, 'r') as f:
        lines = f.readlines()
    lines_atom = None
    natot = sum(nat)    # do not use rin.natot here for EA-vc
    for i, line in enumerate(reversed(lines)):
        if 'ATOMIC_POSITIONS' in line:
            ibegin = len(lines) - 1 - i
            iend = ibegin + natot + 1
            lines_atom = lines[ibegin:iend]
            break
    return lines_atom


def from_lines(lines_cell, lines_atom):
    # ---------- lattice
    unit = lines_cell[0].split()[1]
    if unit[0] == '(' and unit[-1] == ')':
        unit = unit[1:-1]
    if unit.startswith('alat'):
        scale = float(unit[unit.index('=')+1:])    # in Bohr
        scale = float(Length(scale, 'bohr').to('ang'))    # in Ang
    elif unit == 'bohr':
        scale = float(Length(1.0, 'bohr').to('ang'))    # in Ang
    elif unit == 'angstrom':
        scale = 1.0    # in Ang
    else:
        logger.error(f'unit "{unit:s}" for CELL_PARAMETERS'
                         ' is not supported')
    lattice = [[scale * float(x) for x in line.split()]
               for line in lines_cell[1:4]]

    # ---------- species & coordinates
    unit = lines_atom[0].split()[1]
    if unit[0] == '(' and unit[-1] == ')':
        unit = unit[1:-1]
    species = []
    coords = []
    for line in lines_atom[1:]:
        fields = line.split()
        species.append(fields[0])
        coords.append([float(x) for x in fields[1:4]])
        if unit == 'crystal':
            pass    # 'coords' are already internal coordinates
        else:
            logger.error(f'unit "{unit:s}" for ATOMIC_POSITIONS'
                             ' is not supported yet')

    structure = Structure(lattice, species, coords)
    return structure


def write(structure, output, mode='w'):
    # ---------- get in POSCAR format
    poscar = structure.to(fmt='poscar')
    lines = poscar.split('\n')[:-1]
    lines = [line+'\n' for line in lines]

    # ---------- write in QE format
    with open(output, mode) as f:
        f.write('CELL_PARAMETERS angstrom\n')
        for line in lines[2:5]:
            f.write(line)
        f.write('ATOMIC_POSITIONS crystal\n')
        for line in lines[8:]:
            fields = line.split()
            f.write(fields[3]+'  '+' '.join(fields[0:3])+'\n')



'''
Generational change in evolutionary algorithm
'''

from logging import getLogger

import pandas as pd

from .calc_hull import calc_convex_hull_2d
from .ea_child import child_gen
from .survival import survival_fittest
from ..IO import change_input, io_stat, pkl_data
from ..IO.out_results import out_ea_info, out_ea_origin, out_hdist

logger = getLogger('cryspy')


def next_gen(
        rin,
        id_queueing,
        id_running,
        gen,
        init_struc_data,
        opt_struc_data,
        rslt_data,
        ea_vc_data=None,
        struc_mol_id=None,
    ):

    # ---------- log
    logger.info('# ---------- Evolutionary algorithm')
    logger.info(f'Generation {gen + 1}')

    # ---------- current generation
    c_rslt = rslt_data[rslt_data['Gen'] == gen]
    if rin.algo != 'EA-vc':
        c_fitness = c_rslt['E_eV_atom'].to_dict()    # {ID: energy, ...}
    c_ids = c_rslt.index.values    # current IDs [array]

    # ---------- load ea_data. ea_data is used only in this module
    elite_struc = pkl_data.load_elite_struc()
    elite_fitness = pkl_data.load_elite_fitness()
    ea_info = pkl_data.load_ea_info()
    ea_origin = pkl_data.load_ea_origin()

    # ---------- EA-vc
    if rin.algo == 'EA-vc':
        # ------ data for EA-vc
        ef_all = rslt_data['Ef_eV_atom'].to_dict()    # formation energy of all structures
        nat_data, ratio_data, hdist_data = ea_vc_data

        # ------ calc convex hull and hull distance
        hdist = calc_convex_hull_2d(rin, ratio_data, ef_all, c_ids, gen)
        # -- update hdist
        out_hdist(gen, hdist, ratio_data)
        hdist_data[gen] = hdist
        ea_vc_data = (nat_data, ratio_data, hdist_data)    # to use later
        pkl_data.save_hdist_data(hdist_data)

        # ------ fitness (= hull distance) of current generation
        c_fitness = {cid: hdist[cid] for cid in c_ids}
        logger.debug(f'c_fitness: {c_fitness}')

        # ------ update elite_fitness
        #        need to update elite_fitness every time hull distance (convex hull) is updated
        if elite_fitness is not None:
            for cid in elite_fitness:
                elite_fitness[cid] = hdist[cid]
            logger.debug(f'elite_fitness in EA-vc: {elite_fitness}')

    # ---------- survival_fittest
    logger.info('# ------ survival of the fittest')
    c_struc_data = {cid: opt_struc_data[cid] for cid in c_ids}
    ranking, fit_with_elite, struc_wit_elite = survival_fittest(
                                                    c_fitness,
                                                    c_struc_data,
                                                    elite_struc,
                                                    elite_fitness,
                                                    rin.n_fittest,
                                                    rin.fit_reverse,
                                                    rin.emax_ea,
                                                    rin.emin_ea,
                                                )
    logger.info('ranking without duplication (including elite):')
    for cid in ranking:
            logger.info(f'Structure ID {cid:>6}, fitness: {fit_with_elite[cid]:>10.5f}')

    # ---------- generate children by EA
    logger.info('# ------ Generate children')
    # init_struc_data will be updated and  saved in child_gen function
    init_struc_data, parents, operation = child_gen(
                                                rin,
                                                ranking,
                                                fit_with_elite,
                                                struc_wit_elite,
                                                init_struc_data,
                                                struc_mol_id,
                                                ea_vc_data,
                                            )

    # ---------- select elite for next generation
    if rin.n_elite > 0:
        logger.info('# ------ Select elites')
        elite_struc = {}
        elite_fitness = {}
        if rin.algo == 'EA':
            fitness = rslt_data['E_eV_atom'].to_dict()    # {ID: energy, ..,}
            n_elite = rin.n_elite
        if rin.algo == 'EA-vc':
            '''
            In EA-vc, the num. of elite structure is
                the num. of vertices in the convex hull (hull distance < 0.001) + rin.n_elite
            '''
            vert_id = [cid for cid, value in hdist.items() if value < 0.001]
            fitness = hdist
            n_elite = len(vert_id) + rin.n_elite    # temporary
        # ------ ranking for all data
        ranking, _, _ = survival_fittest(fitness, opt_struc_data, None, None,
                                         n_elite, rin.fit_reverse, rin.emax_ea, rin.emin_ea)
        if rin.algo == 'EA-vc':
            '''
            e.g.
            vert_id = [3, 8, 4] <-- assume: structure 3, 8, 4 are identical
            rin.n_elite = 2
            n_elite = len(vert_id) + rin.n_elite = 3 + 2 = 5
            ranking = [3, 6, 9, 11, 15] <-- 8 and 4 are removed because of duplication.
                                            so n_elite should be 1 + 2 = 3.
                        11 and 15 were not originally supposed to be chosen as elite

            Trim the terminals of ranking by the number of duplicated structures in vert_id
            '''
            for cid in vert_id:
                if cid not in ranking:
                    ranking.pop(-1)
        for cid in ranking:
            logger.info(f'Structure ID {cid:>6} keeps as the elite')
            elite_struc[cid] = opt_struc_data[cid]
            elite_fitness[cid] = fitness[cid]
    else:
        elite_struc = None
        elite_fitness = None
    logger.debug(f'elite_fitness for next generation: {elite_fitness}')

    # ---------- new generation
    gen += 1

    # ---------- id_queueing
    id_queueing = [i for i in range(rin.tot_struc, rin.tot_struc + rin.n_pop)]

    # ---------- ea_info
    if rin.algo == 'EA':
        tmp_info = pd.DataFrame(data=[[gen, rin.n_pop, rin.n_crsov, rin.n_perm,
                                       rin.n_strain, rin.n_rand, rin.n_elite,
                                       rin.crs_lat, rin.slct_func]],
                                       columns=ea_info.columns)
    if rin.algo == 'EA-vc':
        tmp_info = pd.DataFrame(data=[[gen, rin.n_pop,
                                    rin.n_crsov, rin.n_perm, rin.n_strain,
                                    rin.n_add, rin.n_elim, rin.n_subs,
                                    rin.n_rand, rin.n_elite,
                                    rin.crs_lat, rin.slct_func]],
                                    columns=ea_info.columns)
    ea_info = pd.concat([ea_info, tmp_info], axis=0, ignore_index=True)
    # ------ out ea_info
    out_ea_info(ea_info)

    # ---------- ea_origin
    # ------ EA operation part
    for cid in range(rin.tot_struc, rin.tot_struc + rin.n_pop - rin.n_rand):
        tmp_origin = pd.DataFrame([[gen, cid, operation[cid],
                                    parents[cid]]], columns=ea_origin.columns)
        ea_origin = pd.concat([ea_origin, tmp_origin], axis=0, ignore_index=True)
    # ------ random part
    for cid in range(rin.tot_struc + rin.n_pop - rin.n_rand,
                     rin.tot_struc + rin.n_pop):
        tmp_origin = pd.DataFrame([[gen, cid, 'random', None]],
                                  columns=ea_origin.columns)
        ea_origin = pd.concat([ea_origin, tmp_origin], axis=0, ignore_index=True)
    # ------ elite part
    if rin.n_elite > 0:
        for cid in ranking:
            tmp_origin = pd.DataFrame([[gen, cid, 'elite', 'elite']],
                                      columns=ea_origin.columns)
            ea_origin = pd.concat([ea_origin, tmp_origin], axis=0, ignore_index=True)

    # ------ out ea_origin
    out_ea_origin(ea_origin)

    # ---------- save ea_id_data
    pkl_data.save_id_queueing(id_queueing)
    pkl_data.save_id_running(id_running)
    pkl_data.save_gen(gen)

    # ---------- save ea_data
    pkl_data.save_elite_struc(elite_struc)
    pkl_data.save_elite_fitness(elite_fitness)
    pkl_data.save_ea_info(ea_info)
    pkl_data.save_ea_origin(ea_origin)

    # ---------- change the value of tot_struc
    next_tot = rin.tot_struc + rin.n_pop
    config = change_input.read_config()
    change_input.change_input(config, 'basic', 'tot_struc', next_tot)
    change_input.write_config(config)
    logger.info('# -- changed cryspy.in')
    logger.info('Changed the value of tot_struc in cryspy.in'
          f' from {rin.tot_struc} to {next_tot}')
    rin.tot_struc = next_tot
    pkl_data.save_input(rin)

    # ---------- status
    stat = io_stat.stat_read()
    io_stat.set_common(stat, 'generation', gen)
    io_stat.set_id(stat, 'id_queueing', id_queueing)
    io_stat.write_stat(stat)

    # ---------- ext
    if rin.calc_code == 'ext':
        with open('ext/stat_job', 'w') as fstat:
            fstat.write('out\n')



from logging import getLogger

import numpy as np
from pymatgen.core import Structure


logger = getLogger('cryspy')


def collect_ase(cid, work_path, nat):
    # ---------- natot
    natot = sum(nat)    # do not use rin.natot here for EA-vc
    # ---------- etc
    magmom = np.nan
    check_opt = 'no_file'    # always no_file in ASE for now
    # ---------- collect energy
    try:
        with open(work_path+'log.tote') as f:
            lines = f.readlines()
        energy = float(lines[-1].split()[0])    # in eV/cell
        energy = energy/float(natot)    # eV/cell --> eV/atom
    except Exception as e:
        energy = np.nan    # error
        logger.warning(f'{e}:    Structure ID {cid}, could not obtain energy from log.tote')
    # ---------- collect CONTCAR
    try:
        opt_struc = Structure.from_file(work_path+'CONTCAR')
    except Exception:
        opt_struc = None
    # ---------- check
    if np.isnan(energy):
        opt_struc = None
    if opt_struc is None:
        energy = np.nan
        magmom = np.nan
    # ---------- return
    return opt_struc, energy, magmom, check_opt


'''
I/O for cryspy.stat
'''
import configparser


def stat_init():
    stat = configparser.ConfigParser()
    stat.add_section('status')
    write_stat(stat)


def stat_read():
    stat = configparser.ConfigParser()
    stat.read('cryspy.stat')
    return stat


def write_stat(stat):
    with open('cryspy.stat', 'w') as f:
        stat.write(f)


# ---------- status section
def set_common(stat, var_str, var):
    stat.set('status', var_str, f'{var}')


def set_id(stat, var_str, var_list):
    if len(var_list) > 30:
        vl = ' '.join(str(a) for a in var_list[:5])
        stat.set('status', var_str, f'{vl} ... total {len(var_list)} IDs')
    else:
        vl = ' '.join(str(a) for a in var_list)
        stat.set('status', var_str, f'{vl}')


def set_stage(stat, cid, cstage):
    stat.set('status', f'ID {cid:>6}', f'Stage {cstage}')


def clean_id(stat, cid):
    stat.remove_option('status', f'ID {cid:>6}')



from logging import getLogger
import random
import numpy as np
from pymatgen.transformations.site_transformations import ReplaceSiteSpeciesTransformation
from ...IO import pkl_data
from ...util.struc_util import get_nat


logger = getLogger('cryspy')


def convex_hull_check(rin):
    '''
    convex hullのどの区間にどれくらいの構造があるかをカウントする
    とりあえずconvex hullを5分割して考える
    ''' 
    # どの区間にどれだけの構造があるかをカウントする
    nat_data, ratio_data = pkl_data.load_ea_vc_data()
    section = [0,0,0,0,0]
    for i in range(len(nat_data)):
        tmp_arr = ratio_data[i][0]
        #logger.info(f'ratio: {tmp_arr}')
        if 0 <= tmp_arr <0.2:
            section[0] +=1
        elif 0.2 <= tmp_arr <0.4:
            section[1] +=1
        elif 0.4 <= tmp_arr <0.6:
            section[2] +=1
        elif 0.6 <= tmp_arr <0.8:
            section[3] +=1
        elif 0.8 <= tmp_arr <=1.0:
            section[4] +=1
        logger.info(f'section: {section}')

    
    if rin.target == 'depop':
        tmp_arr = []
        for i in range(len(section)):
            if i == 0:
                tmp_arr.append(0)
            else:
                if section[tmp_arr[0]] > section[i]:
                    tmp_arr = []
                    tmp_arr.append(i)
                elif section[tmp_arr[0]] == section[i]:
                    tmp_arr.append(i)
        
        if len(tmp_arr) > 1:
            rslt = random.choice(tmp_arr)
        else:
            rslt = tmp_arr[0]
        logger.info(f'tmp_arr: {tmp_arr}')
        logger.info(f'select_section: {rslt}')

    # ---------- 構造が多いところに子構造を生成する場合の処理    
    elif rin.target == 'overpop':
        tmp_arr = []
        for i in range(len(section)):
            if i == 0:
                tmp_arr.append(i)
            else:
                # logger.info(f'i: {i}')
                # logger.info(f'{section[tmp_arr[0]]} < {section[i]}')
                if section[tmp_arr[0]] < section[i]:
                    tmp_arr = []
                    tmp_arr.append(i)
                elif section[tmp_arr[0]] == section[i]:
                    tmp_arr.append(i)
        if len(tmp_arr) > 1:
            rslt = random.choice(tmp_arr)
        else:
            rslt = tmp_arr[0]
        logger.info(f'tmp_arr: {tmp_arr}')
        logger.info(f'select_section: {rslt}')

    return rslt

def operation_atoms(rin, method,child, section):
    """
    構造が少ないところをターゲットにして子構造生成を行う
    methondにはaddition, elimination, substitutionのいずれかを指定する

    section = 0 : 0<=x<0.2
    section = 1 : 0.2<=x<0.4
    section = 2 : 0.4<=x<0.6
    section = 3 : 0.6<=x<0.8
    section = 4 : 0.8<=x<=1.0
    """
    # どの区間に対して子構造を生成するかを決める
    # 候補が多い場合はランダムで一つ決める
    #section = convex_hull_check(rin)
    
    nat,ratio = get_nat(child,rin.atype)
    ratio = ratio[0]
    # ---------- addition
    if method == 'addition':
        logger.info(' ---------- addition')
        cnt = 0
        logger.info(f'section: {section}')
        while True:
            coords = np.random.rand(3)
            if section == 0:
                if 0<= ratio < 0.2: 
                    if cnt == 0:
                        return None
                    else:
                        return child
                elif ratio >= 0.2:
                    child.append(species=rin.atype[1], coords=coords)
                    
            if section == 1:
                if 0.2<= ratio < 0.4:
                    if cnt == 0:
                        return None
                    else:
                        return child
                elif ratio >= 0.4:
                    child.append(species=rin.atype[1], coords=coords)
                    
                else:
                    child.append(species=rin.atype[0], coords=coords)

            if section == 2:
                if 0.4<= ratio < 0.6:
                    if cnt == 0:
                        return None
                    else:
                        return child
                elif ratio >= 0.6:
                    child.append(species=rin.atype[1], coords=coords)
                    
                else:
                    child.append(species=rin.atype[0], coords=coords)
                
            if section == 3:
                if 0.6<= ratio < 0.8:
                    if cnt == 0:
                        return None
                    else:
                        return child
                elif ratio >= 0.8:
                    child.append(species=rin.atype[1], coords=coords)
                    
                else:
                    child.append(species=rin.atype[0], coords=coords)
            
            if section == 4:
                if 0.8<= ratio <= 1.0:
                    if cnt == 0:
                        return None
                    else:
                        return child
                elif ratio >= 1.0:
                    child.append(species=rin.atype[1], coords=coords)
                    
                else:
                    child.append(species=rin.atype[0], coords=coords)

            nat,ratio = get_nat(child,rin.atype)
            ratio = ratio[0]
            cnt += 1
            if cnt == rin.maxcnt_ea:
                #logger.info(f'return None')
                return None


    # ---------- elimination
    elif method == 'elimination':
        logger.info(' ---------- elimination')
        cnt = 0
        logger.info(f'section: {section}')
        # ------ prepare index for each atom type
        indx_each_type = []
        for a in rin.atype:
            indx_each_type.append(
                [i for i, site in enumerate(child)
                    if site.species_string == a])
        while True:
            
            if section == 0:
                if 0<= ratio < 0.2: 
                    if cnt == 0:
                        return None
                    else:
                        return child
                elif ratio >= 0.2:
                    indx = indx_each_type[0]
                    rm = np.random.choice(indx,1,replace=False)
                    child.remove_sites(rm)
                    
            if section == 1:
                if 0.2<= ratio < 0.4:
                    if cnt == 0:
                        return None
                    else:
                        return child
                elif ratio >= 0.4:
                    indx = indx_each_type[0]
                    rm = np.random.choice(indx,1,replace=False)
                    child.remove_sites(rm)
                    
                else:
                    indx = indx_each_type[1]
                    rm = np.random.choice(indx,1,replace=False)
                    child.remove_sites(rm)

            if section == 2:
                if 0.4<= ratio < 0.6:
                    if cnt == 0:
                        return None
                    else:
                        return child
                elif ratio >= 0.6:
                    indx = indx_each_type[0]
                    rm = np.random.choice(indx,1,replace=False)
                    child.remove_sites(rm)
                    
                else:
                    indx = indx_each_type[1]
                    rm = np.random.choice(indx,1,replace=False)
                    child.remove_sites(rm)
                
            if section == 3:
                if 0.6<= ratio < 0.8:
                    if cnt == 0:
                        return None
                    else:
                        return child
                elif ratio >= 0.8:
                    indx = indx_each_type[0]
                    rm = np.random.choice(indx,1,replace=False)
                    child.remove_sites(rm)
                    
                else:
                    indx = indx_each_type[1]
                    rm = np.random.choice(indx,1,replace=False)
                    child.remove_sites(rm)
            
            if section == 4:
                if 0.8<= ratio <= 1.0:
                    if cnt == 0:
                        return None
                    else:
                        return child
                elif ratio >= 1.0:
                    indx = indx_each_type[0]
                    rm = np.random.choice(indx,1,replace=False)
                    child.remove_sites(rm)
                    
                else:
                    indx = indx_each_type[1]
                    rm = np.random.choice(indx,1,replace=False)
                    child.remove_sites(rm)
            from ..ea_child import check_vcnat
            nat_success = check_vcnat(child)
            if nat_success == False:
                return None
            nat,ratio = get_nat(child,rin.atype)
            ratio = ratio[0]
            cnt += 1
            if cnt == rin.maxcnt_ea:
                logger.info(f'return None')
                return None
        
    # ---------- substitution
    else:
        logger.info(' ---------- subsitution')
        cnt = 0
        logger.info(f'section: {section}')
        indx_each_type = []
        for a in rin.atype:
            indx_each_type.append(
                [i for i, site in enumerate(child)
                    if site.species_string == a])
            

        while True:
            if section == 0:
                if 0<= ratio < 0.2: 
                    if cnt == 0:
                        return None
                    else:
                        return child
                elif ratio >= 0.2:
                    indx = indx_each_type[0]
                    rep = np.random.choice(indx,1,replace=False)
                    logger.info(f'rep: {rep}')
                    transform = ReplaceSiteSpeciesTransformation({rep[0]: rin.atype[1]})
                    child = transform.apply_transformation(child)
                    
            if section == 1:
                if 0.2<= ratio < 0.4:
                    if cnt == 0:
                        return None
                    else:
                        return child
                elif ratio >= 0.4:
                    indx = indx_each_type[0]
                    rep = np.random.choice(indx,1,replace=False)
                    #logger.info(f'rep: {rep}')
                    transform = ReplaceSiteSpeciesTransformation({rep[0]: rin.atype[1]})
                    child = transform.apply_transformation(child)
                    
                else:
                    indx = indx_each_type[1]
                    rep = np.random.choice(indx,1,replace=False)
                    #logger.info(f'rep: {rep}')
                    transform = ReplaceSiteSpeciesTransformation({rep[0]: rin.atype[0]})
                    child = transform.apply_transformation(child)

            if section == 2:
                if 0.4<= ratio < 0.6:
                    if cnt == 0:
                        return None
                    else:
                        return child
                elif ratio >= 0.6:
                    indx = indx_each_type[0]
                    rep = np.random.choice(indx,1,replace=False)
                    #logger.info(f'rep: {rep}')
                    transform = ReplaceSiteSpeciesTransformation({rep[0]: rin.atype[1]})
                    child = transform.apply_transformation(child)
                    
                else:
                    indx = indx_each_type[1]
                    rep = np.random.choice(indx,1,replace=False)
                    #logger.info(f'rep: {rep}')
                    transform = ReplaceSiteSpeciesTransformation({rep[0]: rin.atype[0]})
                    child = transform.apply_transformation(child)
                
            if section == 3:
                if 0.6<= ratio < 0.8:
                    if cnt == 0:
                        return None
                    else:
                        return child
                elif ratio >= 0.8:
                    indx = indx_each_type[0]
                    rep = np.random.choice(indx,1,replace=False)
                    #logger.info(f'rep: {rep}')
                    transform = ReplaceSiteSpeciesTransformation({rep[0]: rin.atype[1]})
                    child = transform.apply_transformation(child)
                    
                else:
                    indx = indx_each_type[1]
                    rep = np.random.choice(indx,1,replace=False)
                    #logger.info(f'rep: {rep}')
                    transform = ReplaceSiteSpeciesTransformation({rep[0]: rin.atype[0]})
                    child = transform.apply_transformation(child)
            
            if section == 4:
                if 0.8<= ratio <= 1.0:
                    if cnt == 0:
                        return None
                    else:
                        return child
                elif ratio >= 1.0:
                    indx = indx_each_type[0]
                    rep = np.random.choice(indx,1,replace=False)
                    #logger.info(f'rep: {rep}')
                    transform = ReplaceSiteSpeciesTransformation({rep[0]: rin.atype[1]})
                    child = transform.apply_transformation(child)
                    
                else:
                    indx = indx_each_type[1]
                    rep = np.random.choice(indx,1,replace=False)
                    #logger.info(f'rep: {rep}')
                    transform = ReplaceSiteSpeciesTransformation({rep[0]: rin.atype[0]})
                    child = transform.apply_transformation(child)

            from ..ea_child import check_vcnat
            nat_success = check_vcnat(child)
            if nat_success == False:
                return None
            nat,ratio = get_nat(child,rin.atype)
            ratio = ratio[0]
            cnt += 1
            if cnt == rin.maxcnt_ea:
                #logger.info(f'return None')
                return None








        
    


'''
Structure files in soiap
'''
import itertools

import numpy as np
from pymatgen.core import Structure
from pymatgen.symmetry.analyzer import SpacegroupAnalyzer

from ...util import constants


def from_file(rin, lines, nat):
    # ---------- lattice
    lattice = [[float(x) for x in line.split()] for line in lines[1:4]]
    #     in Bohr, each column is a lattice vector
    lattice = np.array(lattice) * constants.BOHR2ANG  # Bohr --> Ang
    lattice = lattice.T    # each row is a lattice vector

    # ---------- internal coordinates
    coords = [[float(x) for x in line.split()] for line in lines[5:]]

    # ---------- species
    species = [itertools.repeat(typ, times=num) for typ, num in zip(
        rin.atype, nat)]
    species = list(itertools.chain.from_iterable(species))

    structure = Structure(lattice, species, coords)

    return structure


def write(structure, output, symprec=0.01, title="soiap"):
    # ---------- symmetrized structure
    analyzer = SpacegroupAnalyzer(structure, symprec=symprec)
    structure_sym = analyzer.get_symmetrized_structure()

    # ---------- get written data
    lattice = structure_sym.lattice
    symmetry_operations = structure_sym.spacegroup
    site_indices_inequiv = [equiv[0]
                            for equiv in structure_sym.equivalent_indices]
    sites_inequiv = [structure_sym.sites[i] for i in site_indices_inequiv]

    # ---------- write
    with open(output, 'w') as f:
        f.write('data_{0:s}\n'.format(''.join(title.split())))
        f.write('\n')
        for s, value in zip(['a', 'b', 'c'], lattice.abc):
            f.write('{0:17s} {1:12.8f}\n'.format('_cell_length_'+s, value))
        for s, value in zip(['alpha', 'beta', 'gamma'], lattice.angles):
            f.write('{0:17s} {1:12.8f}\n'.format('_cell_angle_'+s, value))
        f.write('\n')
        f.write('loop_\n')
        f.write('  _symmetry_equiv_pos_as_xyz\n')
        for symop in symmetry_operations:
            f.write("  '{0:s}'\n".format(symop.as_xyz_str()))
        f.write('\n')
        f.write('loop_\n')
        f.write('  _atom_site_type_symbol\n')
        for s in ['x', 'y', 'z']:
            f.write('  _atom_site_fract_'+s+'\n')
        for site in sites_inequiv:
            f.write('  {0:5s}  '.format(site.species_string)+' '.join(
                ['{0:9.6f}'.format(x) for x in site.frac_coords])+'\n')



'''
Calculate score in LAQA
'''

import numpy as np


def calc_laqa_bias(force_step, stress_step, wf=0.1, ws=10.0, thres=1.0e-6):
    '''
    force_step: force_step_data at a certain ID, selection (stage)
    stress_step: stress_step_data at a certain ID, selection (stage)

    In this function,
    force_step[step][atom]
    stress_step[step][atom]
    '''
    # ---------- force
    if force_step is None:
        return np.nan
    F1 = np.linalg.norm(force_step[-1], axis=1).mean()
    if len(force_step) == 1:
        F2 = None
    else:
        F2 = np.linalg.norm(force_step[-2], axis=1).mean()
    dF = 1.0 if F2 is None else abs(F1-F2)
    dF = thres if dF < thres else dF

    # ---------- stress
    if stress_step is None:
        return np.nan
    S = np.abs(stress_step[-1]).mean()    # use abs. valule

    # ------ return score
    return wf*F1**2/(2.0*dF) + ws*S    # search maximum


from collections import Counter
from logging import getLogger

import numpy as np
from pymatgen.core import Structure, Lattice
from pymatgen.core.periodic_table import DummySpecie

from ...util.struc_util import origin_shift, sort_by_atype, check_distance
from ...util.struc_util import find_site, cal_g, sort_by_atype_mol, get_nat


logger = getLogger('cryspy')


def gen_crossover(
        atype,
        nat,
        mindist,
        struc_data,
        sp,
        n_crsov,
        id_start=None,
        symprec=0.01,
        crs_lat='random',
        nat_diff_tole=4,
        maxcnt_ea=50,
        vc=False,
        ll_nat=None,
        ul_nat=None,
        struc_mol_id=None,
        molecular=False,
    ):
    '''
    # ---------- args

        tuple may be replaced by list

    atype (tuple): e.g. ('Na', 'Cl')
    nat (tuple): e.g. (4, 4), None if algo == 'EA-vc'
    mindist (tuple): minimum interatomic distance, e.g. ((1.5, 1.5), (1.5, 1.5))
    struc_data (dict): {id: structure data}
    sp (instance): instance of SelectParents class
    n_crsov (int): number of structures to generate by crossover
    id_start (int): start ID for new structures
    symprec (float): tolerance for symmetry
    crs_lat (str): 'equal' or 'random'
    nat_diff_tole (int): tolerance for nat_diff
    maxcnt_ea (int): maximum number of trial in crossover
    vc (bool): set True if algo == 'EA-vc'
    ll_nat (tuple): lower limit of nat for EA-vc, e.g. (1, 1)
    ul_nat (tuple): upper limit of nat for EA-vc, e.g. (8, 8)

    # ---------- return
    children (dict): {id: structure data}
    parents (dict): {id: (id of parent_A, id of parent_B)}
    operation (dict): {id: 'crossover'}
    '''

    # ---------- initialize
    struc_cnt = 0
    children = {}
    #children_mol_id = {}
    parents = {}
    operation = {}

    # ---------- id_offset
    if id_start is None:
        cid = max(struc_data.keys()) + 1
    else:
        if id_start < (max(struc_data.keys()) + 1):
            logger.error('id_start is already included in structure ID of the data')
        else:
            cid = id_start

    # ---------- generate structures by crossover
    while struc_cnt < n_crsov:
        # ------ select parents
        pid_A, pid_B = sp.get_parents(n_parent=2)
        parent_A = struc_data[pid_A]
        parent_B = struc_data[pid_B]
        # ------ generate child
        if molecular:
            logger.error('molecular crossover is not implemented yet.')
            #child, mol_id = co.gen_child_mol(rin, struc_data[pid_A], struc_data[pid_B],
            #                                    struc_mol_id[pid_A], struc_mol_id[pid_B])
        else:
            child = gen_child(atype, nat, mindist, parent_A, parent_B,
                              crs_lat, nat_diff_tole, maxcnt_ea,
                              vc, ll_nat, ul_nat)
        # ------ success
        if child is not None:
            children[cid] = child
            # if molecular:
            #     children_mol_id[cid] = mol_id
            parents[cid] = (pid_A, pid_B)
            operation[cid] = 'crossover'
            try:
                spg_sym, spg_num = child.get_space_group_info(symprec=symprec)
            except TypeError:
                spg_num = 0
                spg_sym = None
            logger.info(f'Structure ID {cid:>6} was generated'
                    f' from {pid_A:>6} and {pid_B:>6} by crossover.'
                    f' Space group: {spg_num:>3} {spg_sym}')
            cid += 1
            struc_cnt += 1

    # ---------- return
    return children, parents, operation


def gen_child(atype, nat, mindist, parent_A, parent_B,
              crs_lat='random', nat_diff_tole=4, maxcnt_ea=50,
              vc=False, ll_nat=None, ul_nat=None):
    '''
    # ---------- args

        tuple may be replaced by list

    atype (tuple): e.g. ('Na', 'Cl')
    nat (tuple): e.g. (4, 4), None if algo == 'EA-vc'
    parent_A (Structure): pymatgen Structure object
    parent_B (Structure): pymatgen Structure object
    mindist (tuple): minimum interatomic distance, e.g. ((1.5, 1.5), (1.5, 1.5))
    crs_lat (str): 'equal' or 'random'
    nat_diff_tole (int): tolerance for nat_diff
    maxcnt_ea (int): maximum number of trial in crossover
    vc (bool): set True if algo == 'EA-vc'
    ll_nat (tuple): lower limit of nat for EA-vc, e.g. (1, 1)
    ul_nat (tuple): upper limit of nat for EA-vc, e.g. (8, 8)

    # ---------- return
    (if success) child (Structure): pymatgen Structure object
    (if fail) None
    '''

    # ---------- initialize
    parent_A = origin_shift(parent_A)    # origin_shift returns a new Structure object
    parent_B = origin_shift(parent_B)
    count = 0

    # ---------- lattice crossover
    if crs_lat == 'equal':
        w_lat = np.array([1.0, 1.0])
    elif crs_lat == 'random':
        w_lat = np.random.choice([0.0, 1.0], size=2, replace=False)
    else:
        logger.error('crs_lat must be equal or random')
    lattice = _lattice_crossover(parent_A, parent_B, w_lat)

    # ---------- generate child
    while True:
        count += 1
        # ------ coordinate crossover
        axis, slice_point, species, coords = _one_point_crossover(parent_A, parent_B)
        # ------ child structure
        child = Structure(lattice, species, coords)
        # ------ check nat_diff
        if not vc:
            nat_diff = _get_nat_diff(atype, nat, child)
            if any([abs(n) > nat_diff_tole for n in nat_diff]):
                logger.debug(f'nat_diff = {nat_diff}')
                if count > maxcnt_ea:    # fail
                    return None
                continue    # slice again
        else:    # EA-vc
            nat_diff = [0, 0]    # dummy
        # ------ check mindist
        success, _, _ = check_distance(child, atype, mindist, check_all=False)
        # ------ something smaller than mindist
        if not success:
            # -- remove atoms within mindist
            if any([n > 0 for n in nat_diff]):
                child = _remove_within_mindist(child, atype, mindist, nat_diff)
                if child is None:    # fail --> slice again
                    if count > maxcnt_ea:
                        return None
                    continue
            else:    # nothing to remove, nat_diff = [0, 0]
                if count > maxcnt_ea:
                    return None
                continue    # fail --> slice again
        if not vc:
            # ------ recheck nat_diff
            # ------ excess of atoms
            nat_diff = _get_nat_diff(atype, nat, child)    # recheck
            if any([n > 0 for n in nat_diff]):
                child = _remove_border_line(child, atype, axis,
                                            slice_point, nat_diff)
            # ------ lack of atoms
            nat_diff = _get_nat_diff(atype, nat, child)    # recheck
            if any([n < 0 for n in nat_diff]):
                child = _add_border_line(child, atype, mindist, axis, slice_point,
                                         nat_diff, maxcnt_ea)
        # ------ nat check for EA-vc
        if vc:
            child_nat, _ = get_nat(child, atype)
            for i, na in enumerate(child_nat):
                if not ll_nat[i] <= na <= ul_nat[i]:
                    logger.warning(f'Crossover: nat = {nat}, ll_nat = {ll_nat}, ul_nat = {ul_nat}')
                    child = None
        # ------ success --> break while loop
        if child is not None:
            break
        # ------ fail --> slice again
        else:
            if count > maxcnt_ea:
                return None
            continue

    # ---------- final check for nat
    if not vc:
        nat_diff = _get_nat_diff(atype, nat, child)
        if not all([n == 0 for n in nat_diff]):
            return None    # failure

    # ---------- sort by atype
    child = sort_by_atype(child, atype)

    # ---------- return
    return child


def _lattice_crossover(parent_A, parent_B, w_lat):
    # ---------- component --> w_lat
    matrix = ((w_lat[0]*parent_A.lattice.matrix
                + w_lat[1]*parent_B.lattice.matrix)
                / w_lat.sum())
    mat_len = np.sqrt((matrix**2).sum(axis=1))
    # ---------- absolute value of vector
    lat_len = ((np.array(parent_A.lattice.abc)*w_lat[0]
                + np.array(parent_B.lattice.abc)*w_lat[1])
                / w_lat.sum())
    # ---------- correction of vector length
    lat_array = np.empty([3, 3])
    for i in range(3):
        lat_array[i] = matrix[i]*lat_len[i]/mat_len[i]
    # ---------- Lattice in pymatgen
    return Lattice(lat_array)


def _one_point_crossover(parent_A, parent_B):
    # ---------- slice point
    while True:
        slice_point = np.random.normal(loc=0.5, scale=0.1)
        if 0.3 <= slice_point <= 0.7:
            break
    axis = np.random.choice([0, 1, 2])

    # ---------- crossover
    species_A = []
    species_B = []
    coords_A = []
    coords_B = []
    for i in range(parent_A.num_sites):
        if parent_A.frac_coords[i, axis] <= slice_point:
            species_A.append(parent_A[i].species_string)
            coords_A.append(parent_A[i].frac_coords)
        else:
            species_B.append(parent_A[i].species_string)
            coords_B.append(parent_A[i].frac_coords)
    for i in range(parent_B.num_sites):
        if parent_B.frac_coords[i, axis] >= slice_point:
            species_A.append(parent_B[i].species_string)
            coords_A.append(parent_B[i].frac_coords)
        else:
            species_B.append(parent_B[i].species_string)
            coords_B.append(parent_B[i].frac_coords)

    # ---------- adopt a structure with more atoms
    if len(species_A) > len(species_B):
        species = species_A
        coords = coords_A
    elif len(species_A) < len(species_B):
        species = species_B
        coords = coords_B
    else:
        if np.random.choice([0, 1]):
            species = species_A
            coords = coords_A
        else:
            species = species_B
            coords = coords_B

    # ---------- return
    return axis, slice_point, species, coords


def _get_nat_diff(atype, nat, child):
    '''
    original nat - child nat
    e.g.
        nat = [4, 4]        # original
        tmp_nat = [3, 5]    # child
        nat_diff = [-1, 1]
    '''
    tmp_nat, _ = get_nat(child, atype)
    nat_diff = [i - j for i, j in zip(tmp_nat, nat)]
    return nat_diff


def _remove_within_mindist(child, atype, mindist, nat_diff):
    '''
    if success: return child
    if fail:    return None
    '''
    for itype in range(len(atype)):
        while nat_diff[itype] > 0:
            # ---------- check dist
            dist_list = check_distance(child, atype, mindist, check_all=True)
            if not dist_list:    # nothing within mindist
                return child
            # ---------- appearance frequency
            ij_within_dist = [isite[0] for isite in dist_list] + [
                jsite[1] for jsite in dist_list]
            site_counter = Counter(ij_within_dist)
            # ---------- get index for removing
            rm_index = None
            # ---- site[0]: index, site[1]: count
            for site in site_counter.most_common():
                if child[site[0]].species_string == atype[itype]:
                    rm_index = site[0]
                    break    # break for loop
            # ---------- remove atom
            if rm_index is None:    # fail
                return None
            else:
                child.remove_sites([rm_index])
                nat_diff[itype] -= 1

    # ---------- final check
    dist_list = check_distance(child, atype, mindist, check_all=True)
    if dist_list:    # still something within mindist
        logger.warning('remove_within_mindist: some atoms within mindist. retry.')
        return None
    else:    # success
        return child


def _remove_border_line(child, atype, axis, slice_point, nat_diff):
    # ---------- rank atoms from border line
    coords_axis = child.frac_coords[:, axis]

    # ---------- boundary --> 0.0, slice_point, 1.0
    near_sp = (slice_point/2.0 < coords_axis) & \
        (coords_axis < (slice_point + 1.0)/2.0)
    near_one = (slice_point + 1.0)/2.0 <= coords_axis

    # ---------- distance from nearest boundary
    coords_diff = np.where(near_sp,
                            abs(coords_axis - slice_point),
                            coords_axis)
    coords_diff = np.where(near_one, 1.0 - coords_diff, coords_diff)
    atom_border_indx = np.argsort(coords_diff)

    # ---------- remove list
    rm_list = []
    for itype, nrm in enumerate(nat_diff):
        rm_list.append([])
        if nrm > 0:
            for ab_indx in atom_border_indx:
                if child[ab_indx].species_string == atype[itype]:
                    rm_list[itype].append(ab_indx)
                if len(rm_list[itype]) == nrm:
                    break

    # ---------- remove
    for each_type in rm_list:
        if each_type:
            child.remove_sites(each_type)

    # ---------- return
    return child


def _add_border_line(child, atype, mindist, axis, slice_point, nat_diff, maxcnt_ea=50):
    for i in range(len(atype)):
        # ---------- counter
        cnt = 0

        # ---------- add atoms
        while nat_diff[i] < 0:
            cnt += 1
            coords = np.random.rand(3)
            mean = _mean_choice(child, axis, slice_point)
            coords[axis] = np.random.normal(loc=mean, scale=0.08)
            child.append(species=atype[i], coords=coords)
            success, mindist_ij, dist = check_distance(child, atype, mindist)
            if success:
                cnt = 0    # reset
                nat_diff[i] += 1
            else:
                type0 = atype[mindist_ij[0]]
                type1 = atype[mindist_ij[1]]
                logger.warning(f'mindist in _add_border_line: {type0} - {type1}, {dist}. retry.')
                child.pop()    # cancel
            # ------ fail
            if cnt == maxcnt_ea:
                return None

        # ---------- return
        return child


def _mean_choice(child, axis, slice_point):
    '''
    Which border contains the most atoms?
    '''
    n_zero = np.sum(np.abs(child.frac_coords[:, axis] - 0.0)
                    < 0.1)
    n_slice = np.sum(np.abs(child.frac_coords[:, axis]
                            - slice_point) < 0.1)
    if n_zero < n_slice:
        mean = 0.0
    elif n_zero > n_slice:
        mean = slice_point
    else:
        mean = np.random.choice([0.0, slice_point])
    return mean



from logging import getLogger
import os

from ..util.utility import check_fwpath
from ..util.struc_util import set_mindist, get_mol_data

# import later
# from .gen_struc_RS import gen_pyxtal
# from ..gen_struc_RS import random_generation


logger = getLogger('cryspy')


def gen_random(rin, nstruc, id_offset, comm, mpi_rank, mpi_size):
    # ---------- log: num of MPI processes
    if mpi_rank == 0:
        logger.info('# ---------- Initial structure generation')
    if mpi_size > 1 and mpi_rank == 0:
        logger.info(f'Number of MPI processes: {mpi_size}')
    # ---------- mindist
    if mpi_rank == 0:
        logger.info('# ------ mindist')
    if mpi_size > 1:
        comm.barrier()
    mindist = set_mindist(rin.atype, rin.mindist, rin.mindist_factor, rin.struc_mode,
                          False, None, mpi_rank)
    if rin.struc_mode == 'mol_bs':
        mindist_dummy = set_mindist(
                            rin.atype,
                            rin.mindist_mol_bs,
                            rin.mindist_mol_bs_factor,
                            rin.struc_mode,
                            dummy=True,
                            mol_file=rin.mol_file,
                            mpi_rank=mpi_rank
                        )

    # ---------- nstruc, offset for MPI
    nstruc_list, offset_list = _divide_task(nstruc, mpi_size, id_offset)

    # ---------- log
    if mpi_rank == 0:
        logger.info('# ------ generate structures')

    # ---------- EA-vc
    vc = True if rin.algo == 'EA-vc' else False

    # ---------- pyxtal
    if not (rin.spgnum == 0 or rin.use_find_wy):
        from .gen_struc_RS import gen_pyxtal
        # ------ crystal
        if rin.struc_mode == 'crystal':
            init_struc_data = gen_pyxtal.gen_struc(
                                  nstruc=nstruc_list[mpi_rank],
                                  atype=rin.atype,
                                  nat=rin.nat,
                                  mindist=mindist,
                                  spgnum=rin.spgnum,
                                  symprec=rin.symprec,
                                  id_offset=offset_list[mpi_rank],
                                  vol_factor=rin.vol_factor,
                                  vol_mu=rin.vol_mu,
                                  vol_sigma=rin.vol_sigma,
                                  vc=vc,
                                  ll_nat=rin.ll_nat,
                                  ul_nat=rin.ul_nat,
                              )
            struc_mol_id = {}     # not used, just for return
        # ------ molecular crystal
        elif rin.struc_mode == 'mol':
            mol_data = get_mol_data(rin.mol_file)
            init_struc_data, struc_mol_id = gen_pyxtal.gen_struc_mol(
                                                nstruc=nstruc_list[mpi_rank],
                                                atype=rin.atype,
                                                nat=rin.nat,
                                                mindist=mindist,
                                                spgnum=rin.spgnum,
                                                mol_data=mol_data,
                                                nmol=rin.nmol,
                                                symprec=rin.symprec,
                                                id_offset=offset_list[mpi_rank],
                                                vol_factor=rin.vol_factor,
                                                vol_mu=rin.vol_mu,
                                                vol_sigma=rin.vol_sigma,
                                                timeout_mol=rin.timeout_mol,
                                            )
        # ------ molecular crystal breaking symmetry
        elif rin.struc_mode == 'mol_bs':
            mol_data = get_mol_data(rin.mol_file)
            init_struc_data, struc_mol_id = gen_pyxtal.gen_struc_mol_break_sym(
                                                nstruc=nstruc_list[mpi_rank],
                                                atype=rin.atype,
                                                nat=rin.nat,
                                                mindist=mindist,
                                                mindist_dummy=mindist_dummy,
                                                spgnum=rin.spgnum,
                                                mol_data=mol_data,
                                                nmol=rin.nmol,
                                                symprec=rin.symprec,
                                                id_offset=offset_list[mpi_rank],
                                                vol_factor=rin.vol_factor,
                                                vol_mu=rin.vol_mu,
                                                vol_sigma=rin.vol_sigma,
                                                rot_mol=rin.rot_mol,
                                                nrot=rin.nrot,
                                            )
    # ---------- w/o pyxtal
    else:
        from .gen_struc_RS import random_generation
        if rin.spgnum == 0:
            init_struc_data = random_generation.gen_wo_spg(
                                    nstruc=nstruc_list[mpi_rank],
                                    atype=rin.atype,
                                    nat=rin.nat,
                                    mindist=mindist,
                                    spgnum=rin.spgnum,
                                    minlen=rin.minlen,
                                    maxlen=rin.maxlen,
                                    dangle=rin.dangle,
                                    symprec=rin.symprec,
                                    maxcnt=rin.maxcnt,
                                    id_offset=offset_list[mpi_rank],
                                    vol_mu=rin.vol_mu,
                                    vol_sigma=rin.vol_sigma,
                                    vc=vc,
                                    ll_nat=rin.ll_nat,
                                    ul_nat=rin.ul_nat,
                                )
            struc_mol_id = {}     # not used, just for return
        else:
            # ---- findwy
            # -- check fwpath
            fwpath = check_fwpath(rin.fwpath)
            if mpi_rank == 0:
                os.makedirs('tmp_gen_struc', exist_ok=True)
            if mpi_size > 1:
                comm.barrier()
            init_struc_data = random_generation.gen_with_find_wy(
                                    nstruc=nstruc_list[mpi_rank],
                                    atype=rin.atype,
                                    nat=rin.nat,
                                    mindist=mindist,
                                    spgnum=rin.spgnum,
                                    minlen=rin.minlen,
                                    maxlen=rin.maxlen,
                                    dangle=rin.dangle,
                                    symprec=rin.symprec,
                                    maxcnt=rin.maxcnt,
                                    id_offset=offset_list[mpi_rank],
                                    vol_mu=rin.vol_mu,
                                    vol_sigma=rin.vol_sigma,
                                    fwpath=fwpath,
                                    mpi_rank=mpi_rank,
                                    vc=vc,
                                    ll_nat=rin.ll_nat,
                                    ul_nat=rin.ul_nat,
                                )
            struc_mol_id = {}     # not used, just for return

    # ------ gather init_struc_data for MPI
    if mpi_size > 1:
        # -- parallel
        # only init_struc_data in rank0 is important
        init_struc_data, struc_mol_id = _gather_struc(
            rin,
            init_struc_data,
            struc_mol_id,
            comm, mpi_rank
        )

    # ---------- return
    # only init_struc_data in rank0 is important
    return init_struc_data, struc_mol_id


def _divide_task(ntask, size, offset=0):
    '''
    # ---------- args
    ntask: int, total number of tasks
    size: int, number of processes
    offset: int, offset of task number

    # ---------- return
    ntask_list: list, number of tasks for each process
    offset_list: list, offset of task number for each process

    # ---------- description
    For example, ntask = 10, size = 4,
    we would like to divide the task as follows:
        rank0 = [0, 1, 2] <-- 3
        rank1 = [3, 4, 5] <-- 3
        rank2 = [6, 7]    <-- 2
        rank3 = [8, 9]    <-- 2

    ntask_list = [3, 3, 2, 2]
    offset_list = [0, 3, 6, 8]

    ntask = 4, size = 1, offset = 5
        rank0 = [5, 6, 7, 8] <-- 4
        ntask_list = [4]
        offset_list = [5]

    case: ntask = 20, size = 4, init_n = 10
        rank0 = [10, 11, 12, 13, 14] <-- 5
        rank1 = [15, 16, 17, 18, 19] <-- 5
        rank2 = [20, 21, 22, 23, 24] <-- 5
        rank3 = [25, 26, 27, 28, 29] <-- 5

    ntask_list = [5, 5, 5, 5]
    offset_list = [10, 15, 20, 25]
    '''
    q = (ntask)//size
    r = (ntask)%size
    ntask_list = []
    offset_list = []
    for i in range(size):
        offset_list.append(offset)
        if r > 0:
            ntask_list.append(q + 1)
            offset += q + 1
            r = r -1
        else:
            ntask_list.append(q)
            offset += q
    return ntask_list, offset_list


def _gather_struc(rin, struc_dict, mol_id_dict, comm, mpi_rank):
    '''
    only if mpi_size > 1
    '''
    # ---------- struc_data
    data = struc_dict
    data = comm.gather(data, root=0)
    if mpi_rank == 0:
        for d in data:
            struc_dict.update(d)
    else:
        assert data is None
    # ---------- for mol_id
    if rin.algo in ['EA', 'EA-vc'] and rin.struc_mode in ['mol', 'mol_bs']:
        data = mol_id_dict
        data = comm.gather(data, root=0)
        if mpi_rank == 0:
            for d in data:
                mol_id_dict.update(d)
        else:
            assert data is None
    # ---------- return
    return struc_dict, mol_id_dict


from logging import getLogger

import numpy as np


logger = getLogger('cryspy')


def calc_ef(energy, ratio, end_point):
    '''
    Formation energy: eV/atom

    note:
    nat = [4, 3, 5]
    12 atoms: energy * 12 - 4 * end_point[0] - 3 * end_point[1] - 5 * end_point[2]
    1 atom: energy - 4/12 * end_point[0] - 3/12 * end_point[1] - 5/12 * end_point[2]
    --> energy - ratio[0] * end_point[0] - ratio[1] * end_point[1] - ratio[2] * end_point[2]
    '''
    # ---------- np.nan
    if np.isnan(energy):
        return np.nan
    # ---------- check
    if len(ratio) != len(end_point):
        logger.error('len(ratio) != len(end_point)')
        raise SystemExit(1)
    # ---------- calc formation energy
    ef = energy
    for x, y in zip(ratio, end_point):
        ef -= x * y
    # ---------- return
    return ef


'''
Append structures by evolutionary algorithm
'''

from logging import getLogger
import os

import pandas as pd

from .ea_child import child_gen
from .survival import survival_fittest
from ..IO import change_input, out_results, pkl_data


logger = getLogger('cryspy')

def append_struc(rin, init_struc_data):
    # ---------- append structures by EA
    logger.info('# ---------- Append structures by EA')

    # ---------- load data
    opt_struc_data = pkl_data.load_opt_struc()
    rslt_data = pkl_data.load_rslt()

    # ---------- fitness
    fitness = rslt_data['E_eV_atom'].to_dict()    # {ID: energy, ..,}

    # ---------- survival_fittest
    logger.info('# ------ survival of the fittest')
    ranking, _, _ = survival_fittest(
                        fitness,
                        opt_struc_data,
                        None,
                        None,
                        rin.n_fittest,
                        rin.fit_reverse,
                        rin.emax_ea,
                        rin.emin_ea,
                    )
    logger.info('ranking without duplication:')
    for cid in ranking:
            logger.info(f'Structure ID {cid:>6}, fitness: {fitness[cid]:>10.5f}')

    # ---------- generate children by EA
    logger.info('# ------ Generate children')
    # init_struc_data will be updated and saved in child_gen function
    init_struc_data, parents, operation = child_gen(
                                                rin,
                                                ranking,
                                                fitness,
                                                opt_struc_data,
                                                init_struc_data,
                                                None,
                                                None,
                                            )

    # ---------- id_queueing
    # id_queueing is treated after this append_struc function

    # ----------  ea_info
    if os.path.isfile('./data/pkl_data/ea_info.pkl'):
        ea_info = pkl_data.load_ea_info()
        ea_origin = pkl_data.load_ea_origin()
    else:
        # ------ initialize
        # -- ea_info
        ea_info = pd.DataFrame(columns=['Gen',
                                        'Population',
                                        'Crossover',
                                        'Permutation',
                                        'Strain',
                                        'Random',
                                        'Elite',
                                        'crs_lat',
                                        'slct_func',
                                        ])
        ea_info.iloc[:, 0:7] = ea_info.iloc[:, 0:7].astype(int)
        # -- ea_origin
        ea_origin = pd.DataFrame(columns=['Gen', 'Struc_ID', 'Operation', 'Parent'])
        ea_origin.iloc[:, 0:2] = ea_origin.iloc[:, 0:2].astype(int)
    # ------ register ea_info
    tmp_info = pd.DataFrame([[
                        rin.tot_struc,
                        rin.n_pop,
                        rin.n_crsov,
                        rin.n_perm,
                        rin.n_strain,
                        rin.n_rand,
                        0,
                        rin.crs_lat,
                        rin.slct_func
                    ]],
                    columns=ea_info.columns
                )
    ea_info = pd.concat([ea_info, tmp_info], axis=0, ignore_index=True)
    # ------ out ea_info
    out_results.out_ea_info(ea_info)

    # ---------- ea_origin
    # ------ EA operation part
    for cid in range(rin.tot_struc, rin.tot_struc + rin.n_pop - rin.n_rand):
        tmp_origin = pd.DataFrame([[
                            rin.tot_struc,
                            cid, operation[cid],
                            parents[cid],
                        ]],
                        columns=ea_origin.columns
                    )
        ea_origin = pd.concat([ea_origin, tmp_origin], axis=0, ignore_index=True)
    # ------ random part
    for cid in range(rin.tot_struc + rin.n_pop - rin.n_rand, rin.tot_struc + rin.n_pop):
        tmp_origin = pd.DataFrame([[
                            rin.tot_struc,
                            cid,
                            'random',
                            None
                        ]],
                        columns=ea_origin.columns
                    )
        ea_origin = pd.concat([ea_origin, tmp_origin], axis=0, ignore_index=True)
    # ------  out ea_origin
    out_results.out_ea_origin(ea_origin)

    # ---------- save ea_data
    pkl_data.save_ea_info(ea_info)
    pkl_data.save_ea_origin(ea_origin)

    # ---------- change variables in cryspy.in
    next_tot = rin.tot_struc + rin.n_pop
    config = change_input.read_config()
    logger.info('# -- Changed cryspy.in')
    # ------ tot_struc
    change_input.change_input(config, 'basic', 'tot_struc', next_tot)
    logger.info(f'Changed the value of tot_struc in cryspy.in from {rin.tot_struc} to {next_tot}')
    rin.tot_struc = next_tot
    # ------ append_struc_ea: True --> False
    change_input.change_input(config, 'option', 'append_struc_ea', False)
    logger.info('Changed the value of append_struc_ea in cryspy.in from True to Flase')
    # ------ write and save
    change_input.write_config(config)
    pkl_data.save_input(rin)

    # ---------- return
    return init_struc_data
